#include <config.h>

!#define _MD_STATE_DEBUG_

module md_status_class
!
!= Defines Molecular Dynamics algorithm/controller status class
!
!== Version
!
! $Id: md_status_class.F90,v 1.21 2011/09/25 14:04:23 ishikawa Exp $
!
  use constants_module
  use comlib
  use lattice_class, only : nodeid, CHARLEN
  use error_class
  implicit none
  private

  public :: md_status
  public :: new
  public :: delete
  public :: read
  public :: save
  public :: print
  public :: get_dt
  public :: get_depth
  public :: get_parameter_string
  public :: is_p
  public :: is_loop_end
  public :: set_time_step_forward
  public :: set_time_step_backward

  integer, parameter, public :: MD_QPQ_MTS_SIMPLE = +1
  integer, parameter, public :: MD_PQP_MTS_SIMPLE = -1
  integer, parameter, public :: MD_QPQ_STS_IMPROVED = +2
  integer, parameter, public :: MD_PQP_STS_IMPROVED = -2
  integer, parameter, public :: MD_QPQ_MTS_OMELYAN = +3
  integer, parameter, public :: MD_PQP_MTS_OMELYAN = -3

  real(DP), parameter :: OMELYAN_LAMBDA = 0.193183_DP    ! Omelyan integrator parameter constant


  character(*), parameter :: ALGNAME(-3:3) = &
 &         (/"      Multiple Time Step, Omelyan PQPQP :",   & ! -3
 &           "         Single Time Step, Improved PQP :",   & ! -2
 &           "Multiple Time Step, Simple leapforg PQP :",   & ! -1
 &           "                                    NON :",   & !  0
 &           "Multiple Time Step, Simple leapforg QPQ :",   & ! +1
 &           "         Single Time Step, Improved QPQ :",   & ! +2
 &           "      Multiple Time Step, Omelyan QPQPQ :"/)    ! +3 
  ! algorithm name strings

  type md_state
  !
  ! MD step state
  !
    real(DP) :: dt     = 0.0_DP
    integer  :: depth  = 0
    logical  :: is_p   = .FALSE.
  end type

  type md_status
  !
  ! MD controller status
  !
    private
    type(md_state), pointer :: state(:) => NULL()
    integer,        pointer :: nmd(:)   => NULL()
    real(DP),       pointer :: tau_p(:) => NULL()
    real(DP) :: tau   = 0.0_DP
    real(DP) :: tau_q = 0.0_DP
    logical :: trial_run = .TRUE.
    integer :: algorithm  = 0
    integer :: max_depth  = 1
    integer :: step_count = 0
    integer :: total_steps = 0
    integer :: current_step = 0
    integer :: reverse_time = 0
  end type

  interface new
    module procedure new_md_status
  end interface

  interface delete
    module procedure delete_md_status
  end interface

  interface setup
    module procedure setup_md_status
  end interface

  interface read
    module procedure read_md_status
  end interface

  interface save
    module procedure save_md_status
  end interface

  interface print
    module procedure print_md_status
  end interface

  interface get_dt
    module procedure get_dt_md_status
  end interface

  interface get_depth
    module procedure get_depth_md_status
  end interface

  interface is_p
    module procedure is_p_md_status
  end interface

  interface is_loop_end
    module procedure is_loop_end_md_status
  end interface

  interface get_parameter_string
    module procedure get_parameter_string_md_status
  end interface

contains

subroutine set_time_step_backward(this)
!
! set MD controller to do backward integration
!
  implicit none
  type(md_status), intent(inout) :: this
  this%reverse_time = 1
  return
end subroutine

subroutine set_time_step_forward(this)
!
! set MD controller to do forward integration
!
  implicit none
  type(md_status), intent(inout) :: this
  this%reverse_time = 0
  return
end subroutine

function get_dt_md_status(this) result(dt)
!
! return MD step size dt for current step.
!
  implicit none
  type(md_status), intent(in) :: this
  real(DP) :: dt
  integer :: i
  i = this%current_step
  if (1==this%reverse_time) then
    dt = -this%state(i)%dt
  else
    dt =  this%state(i)%dt
  endif
  return
end function

function get_depth_md_status(this) result(depth)
!
! return MD step depth for current step.
!
  implicit none
  type(md_status), intent(in) :: this
  integer :: depth
  integer :: i
  i = this%current_step
  depth = this%state(i)%depth
  return
end function

function is_p_md_status(this) result(is_p)
!
! return MD update type for current step.
!
! .true.:: user should update momentum (P)
! .false.:: user should update corrdinate (Q)
!
  implicit none
  type(md_status), intent(in) :: this
  logical :: is_p
  integer :: i
  i = this%current_step
  is_p = this%state(i)%is_p
  return
end function

function is_loop_end_md_status(this) result(is_end)
!
! return MD controller loop status
!
! .true.:: MD required loop counts are reached. exit MD step loop.
! .false.:: MD step loop still remains.  go next loop step.
!
  implicit none
  type(md_status), intent(inout) :: this
  logical :: is_end
  integer :: i
  i = this%current_step
  if (i == this%total_steps) then
    is_end = .TRUE.
    this%current_step = 0
  else
    is_end = .FALSE.
    this%current_step = this%current_step + 1
  endif
  return
end function

function get_parameter_string_md_status(this) result(str)
!
! return MD controller parameters in formatted string.
!
  implicit none
  type(md_status), intent(in) :: this
  character(CHARLEN) :: str
  integer :: i
  write(str,'(A," tau =",E22.15," NMD =",10I3)') &
 &  TRIM(ADJUSTL(ALGNAME(this%algorithm))),this%tau,(this%nmd(i),i=this%max_depth,1,-1)
  return
end function

subroutine new_md_status(this)
!
! initialize/create MD controller
!
  implicit none
  type(md_status), intent(inout) :: this
  this%state => NULL()
  this%nmd   => NULL()
  this%tau_p => NULL()
  this%tau   = 0.0_DP
  this%tau_q = 0.0_DP
  this%algorithm    = 0
  this%max_depth    = 1
  this%step_count   = 0
  this%total_steps  = 0
  this%current_step = 0
  this%reverse_time = 0
  return
end subroutine

subroutine delete_md_status(this)
!
! delete MD controller
!
  implicit none
  type(md_status), intent(inout) :: this
  if (associated(this%state)) deallocate(this%state)
  if (associated(this%nmd))   deallocate(this%nmd)
  if (associated(this%tau_p)) deallocate(this%tau_p)
  return
end subroutine

subroutine read_md_status(this,iout)
!
! \Read MD controller parameters from formatted file (by unit number)
!
  implicit none
  type(md_status), intent(inout) :: this
  integer, intent(in) :: iout
  integer :: i
  if (nodeid == 0) then
    read(iout,*)this%algorithm
    read(iout,*)this%tau
    read(iout,*)this%max_depth
  endif
#ifndef _singlePU
  call comlib_bcast(this%algorithm,0)
  call comlib_bcast(this%tau,0)
  call comlib_bcast(this%max_depth,0)
#endif
  allocate(this%nmd(1:this%max_depth))
  if (nodeid == 0) then
    read(iout,*)(this%nmd(i),i=this%max_depth,1,-1)
  endif
#ifndef _singlePU
  do i=1,this%max_depth
    call comlib_bcast(this%nmd(i),0)
  enddo
#endif

  call setup_md_status(this)

  return
end subroutine

subroutine save_md_status(this,iout)
!
! \Save MD controller parameters to formatted file (by unit number)
!
  implicit none
  type(md_status), intent(inout) :: this
  integer,intent(in) :: iout
  integer :: i
  if (nodeid == 0) then
    write(iout,*)         this%algorithm
    write(iout,'(F18.12)')this%tau
    write(iout,*)         this%max_depth
    write(iout,*)(this%nmd(i),i=this%max_depth,1,-1)
  endif
  return
end subroutine


subroutine print_md_status(this)
!
! \Print out MD controller parameters on display
!
  use comlib
  implicit none
  type(md_status), intent(inout) :: this
  real(DP) :: dt
  integer :: i,iflg

  iflg=0
  if (nodeid==0) then
    select case(this%algorithm)
    case (MD_QPQ_MTS_SIMPLE)
      write(*,'(8X,"        MD Integ method : Multiple time scale Simple leapflog (QPQ order)")')
    case (MD_PQP_MTS_SIMPLE)
      write(*,'(8X,"        MD Integ method : Multiple time scale Simple leapflog (PQP order)")')
    case (MD_QPQ_STS_IMPROVED)
      write(*,'(8X,"        MD Integ method : Single time scale Improved leapflog (QPQ order)")')
      if (this%max_depth > 1) then
        write(*,'(8X,"        Error in max_detph !")')
        iflg=1
      endif
    case (MD_PQP_STS_IMPROVED)
      write(*,'(8X,"        MD Integ method : Single time scale Improved leapflog (PQP order)")')
      if (this%max_depth > 1) then
        write(*,'(8X,"        Error in max_detph !")')
        iflg=1
      endif
    case (MD_QPQ_MTS_OMELYAN)
      write(*,'(8X,"        MD Integ method : Multiple time scale Omelyan (QPQPQ order)")')
      write(*,'(8X,"                          with Lambda = ",F14.8)')OMELYAN_LAMBDA
    case (MD_PQP_MTS_OMELYAN)
      write(*,'(8X,"        MD Integ method : Multiple time scale Omelyan (PQPQP order)")')
      write(*,'(8X,"                          with Lambda = ",F14.8)')OMELYAN_LAMBDA
    end select
    write(*,*)
    write(*,'(6X,"        Trajectory Length :",E24.16)')this%tau
    write(*,'(6X,"         Time Scale Depth :",I3)')    this%max_depth
    dt = this%tau
    do i=this%max_depth,1,-1
      dt=dt/this%nmd(i)
      write(*,'(8X,"       MD step  dt(",I3,") :",E24.16)')i,dt
      write(*,'(8X,"  # of MD step Nmd(",I3,") :",I4)')    i,this%nmd(i)
    enddo
    write(*,*)
  endif

#ifndef _singlePU
  call comlib_bcast(iflg,0)
#endif

  if (iflg==1) then
    call error_stop("Error in Inputfile.")
  endif

  return
end subroutine

subroutine setup_md_status(this)
!
! Initialize MD controller. this is called in read method.
!
  implicit none
  type(md_status), intent(inout) :: this
  integer :: nstep, ndepth
  real(DP) :: tau
  logical :: is_ok

  is_ok = .TRUE.
  if (this%max_depth <= 0) then
    is_ok = .FALSE.
  endif
  if (.not.associated(this%nmd)) then
    is_ok = .FALSE.
  endif

  if (.not. is_ok) then
    call error_stop("md status setup failed.")
  endif

  if (.not.associated(this%tau_p)) then
    allocate(this%tau_p(this%max_depth))
  endif

  if (.not.associated(this%state)) then
    allocate(this%state(this%max_depth))
  endif

     tau = this%tau
  ndepth = this%max_depth
  nstep  = this%nmd(ndepth)

  this%trial_run = .TRUE.

  select case (this%algorithm)
  case(MD_QPQ_MTS_SIMPLE)

    call md_leapfrog_qpq(this,nstep,ndepth,tau) ! total step count survey loop
    call md_create(this)                        ! allocate memory for each step
    call md_leapfrog_qpq(this,nstep,ndepth,tau) ! set status for each step

  case(MD_PQP_MTS_SIMPLE)

    call md_leapfrog_pqp(this,nstep,ndepth,tau)
    call md_create(this)
    call md_leapfrog_pqp(this,nstep,ndepth,tau)

  case(MD_QPQ_MTS_OMELYAN)

    call md_omelyan_qpq(this,nstep,ndepth,tau)
    call md_create(this)
    call md_omelyan_qpq(this,nstep,ndepth,tau)

  case(MD_PQP_MTS_OMELYAN)

    call md_omelyan_pqp(this,nstep,ndepth,tau)
    call md_create(this)
    call md_omelyan_pqp(this,nstep,ndepth,tau)

  case default
    call error_stop("Other integrator is not yet implemented.")
  end select

  return
end subroutine

subroutine update_q(this,idepth,dt)
!
! set update q state for initializing MD controller.
!
! this is used in 
!
  implicit none
  type(md_status), intent(inout) :: this
  integer, intent(in) :: idepth
  real(DP), intent(in) :: dt
  integer :: i

  this%step_count = this%step_count + 1

  if (.not. this%trial_run) then
    i = this%step_count
    this%tau_q = this%tau_q + dt
    this%state(i)%dt = dt
    this%state(i)%depth = idepth
    this%state(i)%is_p = .FALSE.
#ifdef _MD_STATE_DEBUG_
    if (nodeid==0) &
 &   write(*,'("update_q : depth=",I3," dt=",E24.15," tau_q=",E24.15)')idepth,dt,this%tau_q
#endif
  endif

  return
end subroutine

subroutine update_p(this,idepth,dt)
!
! set update p state for initializing MD controller.
!
! this is used in 
!
  implicit none
  type(md_status), intent(inout) :: this
  integer, intent(in) :: idepth
  real(DP), intent(in) :: dt
  integer :: i

  this%step_count = this%step_count + 1

  if (.not. this%trial_run) then
    i = this%step_count
    this%tau_p(idepth) = this%tau_p(idepth) + dt
    this%state(i)%dt = dt
    this%state(i)%depth = idepth
    this%state(i)%is_p = .TRUE.

#ifdef _MD_STATE_DEBUG_
    if (nodeid==0) &
 &   write(*,'("update_p : depth=",I3," dt=",E24.15," tau_p=",E24.15)')idepth,dt,this%tau_p(idepth)
#endif
  endif

  return
end subroutine

subroutine md_create(this)
!
! state memory allocation for initializaton
!
  implicit none
  type(md_status), intent(inout) :: this

  if (this%trial_run) then

    if (0 == this%step_count) then
      call error_stop("Do trial run before calling md_create.")
    endif

    this%total_steps = this%step_count

    if (associated(this%state)) then
      deallocate(this%state)
    endif

    allocate(this%state(this%total_steps))
    allocate(this%tau_p(this%total_steps))
    this%tau_p(:) = 0.0_DP
    this%step_count = 0

    this%trial_run = .FALSE.
  endif

  return
end subroutine

recursive subroutine md_leapfrog_qpq(this,current_nstep,current_depth,current_tau)
!
! Molecular dynamics integrator initializer for Simple leapfrog scheme in QPQ order
!
! This is recursively defined.
!
  use comlib
  implicit none
  type(md_status), intent(inout) :: this ! MD order/state to be set
  integer,  intent(in) :: current_nstep  ! current number of steps
  integer,  intent(in) :: current_depth  ! current MD depth
  real(DP), intent(in) :: current_tau    ! current MD length

  real(DP) :: dt,dt2
  integer :: imd,next_depth,next_nstep,next_nstep2
  character(CHARLEN) :: str
  

  dt  = current_tau/current_nstep
  dt2 = dt/2.0_DP

  if (1 /= current_depth) then

    next_depth = current_depth-1
    next_nstep = this%nmd(next_depth) ! get nubmer of steps for next depth 
  
    if (SIZE(this%nmd) > 1 .and. current_depth /= SIZE(this%nmd) ) then
      if (mod(next_nstep,2) /= 0) then
        write(str,'("Multiplie timestep MD integrator (QPQ order)")')
        call error_message(TRIM(str))
        write(str,'("The MD step number should be even. NMD(",I2,")=",I4)')next_depth,next_nstep
        call error_stop(TRIM(str))
      endif
    endif
    next_nstep2 = next_nstep/2
  

    call md_leapfrog_qpq(this,next_nstep2,next_depth,dt2)
    do imd = 1,current_nstep-1
      call update_p(this,current_depth,dt)
      call md_leapfrog_qpq(this,next_nstep,next_depth,dt)
    enddo
    call update_p(this,current_depth,dt)
    call md_leapfrog_qpq(this,next_nstep2,next_depth,dt2)

  else 

    call update_q(this,current_depth,dt2)
    do imd = 1,current_nstep-1
      call update_p(this,current_depth,dt)
      call update_q(this,current_depth,dt)
    enddo
    call update_p(this,current_depth,dt)
    call update_q(this,current_depth,dt2)

  endif

  return
end subroutine

recursive subroutine md_leapfrog_pqp(this,current_nstep,current_depth,current_tau)
!
! Molecular dynamics integrator initializer for Simple leapfrog scheme in PQP order
!
! This is recursively defined.
!
  use comlib
  implicit none
  type(md_status), intent(inout) :: this
  integer,  intent(in) :: current_nstep
  integer,  intent(in) :: current_depth
  real(DP), intent(in) :: current_tau

  real(DP) :: dt,dt2
  integer :: imd,next_depth,next_nstep
  

  dt  = current_tau/current_nstep
  dt2 = dt/2.0_DP
  
  if (1 /= current_depth) then

    next_depth = current_depth-1
    next_nstep = this%nmd(next_depth) ! get nubmer of steps for next depth 

    call update_p(this,current_depth,dt2)
    do imd = 1,current_nstep-1
      call md_leapfrog_qpq(this,next_nstep,next_depth,dt)
      call update_p(this,current_depth,dt)
    enddo
    call md_leapfrog_qpq(this,next_nstep,next_depth,dt)
    call update_p(this,current_depth,dt2)

  else 

    call update_p(this,current_depth,dt2)
    do imd = 1,current_nstep-1
      call update_q(this,current_depth,dt)
      call update_p(this,current_depth,dt)
    enddo
    call update_q(this,current_depth,dt)
    call update_p(this,current_depth,dt2)

  endif

  return
end subroutine

recursive subroutine md_omelyan_qpq(this,current_nstep,current_depth,current_tau)
!
! Molecular dynamics integrator initializer for Omelyan scheme in QPQPQ order
!
! This is recursively defined.
!
  use comlib
  implicit none
  type(md_status), intent(inout) :: this
  integer,  intent(in) :: current_nstep
  integer,  intent(in) :: current_depth
  real(DP), intent(in) :: current_tau

  real(DP) :: dt,dt2,OLmdt,MLmdt
  integer :: imd,next_depth,next_nstep,next_nstep_out,next_nstep_mid
  
  dt    = current_tau/current_nstep
  dt2   = dt/2.0_DP
  OLmdt = OMELYAN_LAMBDA*dt
  MLmdt = (1.0_DP - 2*OMELYAN_LAMBDA)*dt

  if (1 /= current_depth) then

    next_depth = current_depth-1
    next_nstep = this%nmd(next_depth) ! get nubmer of steps for next depth 
    next_nstep_out = CEILING(OMELYAN_LAMBDA*real(next_nstep,KIND=DP))
    next_nstep_mid = CEILING((1.0_DP-2*OMELYAN_LAMBDA)*real(next_nstep,KIND=DP))

    call md_omelyan_qpq(this,next_nstep_out,next_depth,OLmdt)
    do imd = 1,current_nstep-1
      call update_p(this,current_depth,dt2)
      call md_omelyan_qpq(this,next_nstep_mid,next_depth,MLmdt)
      call update_p(this,current_depth,dt2)
      call md_omelyan_qpq(this,next_nstep_out*2,next_depth,OLmdt*2)
    enddo
    call update_p(this,current_depth,dt2)
    call md_omelyan_qpq(this,next_nstep_mid,next_depth,MLmdt)
    call update_p(this,current_depth,dt2)
    call md_omelyan_qpq(this,next_nstep_out,next_depth,OLmdt)

  else 

    call update_q(this,current_depth,OLmdt)
    do imd = 1,current_nstep-1
      call update_p(this,current_depth,dt2)
      call update_q(this,current_depth,MLmdt)
      call update_p(this,current_depth,dt2)
      call update_q(this,current_depth,OLmdt*2)
    enddo
    call update_p(this,current_depth,dt2)
    call update_q(this,current_depth,MLmdt)
    call update_p(this,current_depth,dt2)
    call update_q(this,current_depth,OLmdt)

  endif

  return
end subroutine

recursive subroutine md_omelyan_pqp(this,current_nstep,current_depth,current_tau)
!
! Molecular dynamics integrator initializer for Omelyan scheme in PQPQP order
!
! This is recursively defined.
!
  use comlib
  implicit none
  type(md_status), intent(inout) :: this
  integer,  intent(in) :: current_nstep
  integer,  intent(in) :: current_depth
  real(DP), intent(in) :: current_tau

  real(DP) :: dt,dt2,Lmdt,MLmdt
  integer :: imd,next_depth,next_nstep
  
  dt    = current_tau/current_nstep
  dt2   = dt/2.0_DP
   Lmdt = OMELYAN_LAMBDA*dt
  MLmdt = (1.0_DP - 2*OMELYAN_LAMBDA)*dt

  if (1 /= current_depth) then

    next_depth = current_depth-1
    next_nstep = this%nmd(next_depth) ! get nubmer of steps for next depth 

    call update_p(this,current_depth,Lmdt)
    do imd = 1,current_nstep-1
      call md_omelyan_pqp(this,next_nstep,next_depth,dt2)
      call update_p(this,current_depth,MLmdt)
      call md_omelyan_pqp(this,next_nstep,next_depth,dt2)
      call update_p(this,current_depth,Lmdt*2)
    enddo
    call md_omelyan_pqp(this,next_nstep,next_depth,dt2)
    call update_p(this,current_depth,MLmdt)
    call md_omelyan_pqp(this,next_nstep,next_depth,dt2)
    call update_p(this,current_depth,Lmdt)

  else 

    call update_p(this,current_depth,Lmdt)
    do imd = 1,current_nstep-1
      call update_q(this,current_depth,dt2)
      call update_p(this,current_depth,MLmdt)
      call update_q(this,current_depth,dt2)
      call update_p(this,current_depth,Lmdt*2)
    enddo
    call update_q(this,current_depth,dt2)
    call update_p(this,current_depth,MLmdt)
    call update_q(this,current_depth,dt2)
    call update_p(this,current_depth,Lmdt)

  endif

  return
end subroutine

end module
