#include <config.h>

!#define _DEBUG
!#undef _DEBUG

!
! If _PREC_EO_ defined, the 4D even/odd site preconditioned solver is used
! to invert domainwall operators.
!
#define _PREC_NON_     (1)
#define _PREC_EO_      (2)
#define _PREC_4DSSOR_  (3)
#define _PREC_ _PREC_EO_
!#define _PREC_ _PREC_4DSSOR_

module quark_dwf_class
!
!= Defines Domainwall fermion
!
! Note chiral projection notation
!
! $ R y = (1-\gamma_5) y = P^{-} y $ , $ L y = (1+\gamma_5) y = P^{+} y $ 
!
! Domainwall operator notation:
!
!\[
!  (D_{DW}(m_q))(n,m)_{s,r}  =  (5-M) E(n,m) \otimes E_{s,r}
!                - \frac{1}{2} M_{hop}(n,m) \otimes E_{s,r} \] \[
!           - E(n,m) \otimes \frac{1}{2}
!                             \left[ (1-\gamma_5) \delta_{s+1,r}
!                                  + (1+\gamma_5) \delta_{s-1,r} \right.  \] \[ \left.
!                              - m_q (1-\gamma_5) \delta_{s,N}\delta_{1,r} 
!                              - m_q (1+\gamma_5) \delta_{s,1}\delta_{N,r} \right]
!\]
!
! Hoping operator notation:
!
!\[
!  M_{hop}(n,m) = \sum_{\mu=1}^{4}
!      \left[ (1-\gamma_{\mu})U_{\mu}(n)\delta_{n+\hat{\mu},m}
!           + (1+\gamma_{\mu})U_{\mu}^{\dag}(m)\delta_{n-\hat{\mu},m} \right]
!\]
!
!== Version
!
! $Id: quark_dwf_class.F90,v 1.24 2011/06/14 11:10:33 ishikawa Exp $
!
  use comlib
  use lattice_class
  use action_base_class
  use timer_class
  use error_class
  use solver_parameter_class
  use chrolog_class
  use hmc_status_class
  use field_fermion_class
  use field_gauge_class
  use quark_wilson_class

  use field_5dfermion_class

  use perfmon_class
  implicit none
  private

  integer, parameter :: NGX=2
  integer, parameter :: NGY=2
  integer, parameter :: NGZ=1
  integer, parameter :: NBX=NX/NGX
  integer, parameter :: NBY=NY/NGY
  integer, parameter :: NBZ=NZ/NGZ
  integer, parameter :: NOMP=NGX*NGY*NGZ
#ifdef _PERFMON_
  type(flop_counter), save :: m_flop
#endif

  public :: CRON_NSITE
#if _PREC_ == _PREC_EO_
  integer, parameter :: CRON_NSITE=NHSITE
#else
  integer, parameter :: CRON_NSITE=NSITE
#endif

  public :: DW_TYPE_SHAMIR
  public :: DW_TYPE_BORICI
  public :: DW_TYPE_ZOLOTAREV_SHAMIR
  public :: DW_TYPE_ZOLOTAREV_CHIU
  integer, parameter :: DW_TYPE_SHAMIR           = 1  ! standard domain-wall fermion (Shamir kernel)
  integer, parameter :: DW_TYPE_BORICI           = 2  !   Borici domain-wall fermion (Overlap kernel)
  integer, parameter :: DW_TYPE_ZOLOTAREV_SHAMIR = 3  ! Zolotarev improved fermion (Overlap with Shamir kernel)
  integer, parameter :: DW_TYPE_ZOLOTAREV_CHIU   = 4  ! Zolotarev improved fermion (Overlap, Chiu)
  character(len=CHARLEN), parameter :: DW_TYPE_NAME(4)= &
 &           [character(len=CHARLEN):: " Shamir (standared DWF)",&
 &                                     " Borici (overlap DWF)",  &
 &                                     " Zolotarev Shamir (Shamir kernel)",&
 &                                     " Zolotarev Chiu  (overlap kernel)"]

  character(len=*), parameter :: ACTION_NAME = "Domainwall Quark"

  public :: mult_iter
  public :: copy_fq_time

  public :: su3fv_spinor
  public :: field_quark_eo_wg
  public :: field_quark_eo_wog
  public :: field_quark_wg
  public :: field_quark_wog
  public :: set_gaussian_noise
  public :: mult_hopping_tzyx_eo 

  public :: field_dw_quark_wg
  public :: pack
  public :: unpack
  public :: abs2
  public :: prod
  public :: clear
  public :: assign
  public :: assign_add
  public :: accum_add
  public :: assign_sub
  public :: accum_sub
  public :: accum_mult
  public :: copy_boundary 
  public :: force_hmc_hopping

  public :: quark_domainwall
  public :: make_mass_term
  public :: set_id
  public :: get_id
  public :: get_NS
  public :: get_mass
  public :: set_mass
  public :: set_coef
  public :: new
  public :: delete
  public :: read
  public :: print
  public :: save_config
  public :: assign_mult_gamma5
  public :: accum_mult_gamma5
  public :: assign_proj_5dto4d
  public :: assign_proj_4dto5d
  public :: assign_mult_impterm_dwf
  public :: assign_mult_dwf
  public :: assign_inv_mult_dwf

  public :: dwf_low_modes
  public :: get_low_modes

!========================================================
!
! Define Lowmode containers for Domainwall fermion for chiral-symmetry improvement
!
!========================================================

  type :: dwf_low_modes
  !
  ! low modes continer for domain wall operator
  !
    integer :: NEV
    type(field_dw_quark_wg), private, allocatable :: W
    complex(DP),             private, allocatable :: XX(:,:)
  end type

  interface new
  !
  ! create dwf low-mode container
  !
    module procedure new_dwq_low_modes
  end interface

  interface delete
  !
  ! create dwf low-mode container
  !
    module procedure delete_dwq_low_modes
  end interface

  interface get_low_modes
  !
  ! compute low modes for chiral-symmetry improvement
  !
    module procedure get_dwq_low_modes
  end interface


!========================================================
!
! Define Domainwall fermion paraemters and operators
!
!========================================================
  type, extends(quark_wilson) :: quark_domainwall
  !
  ! Domainwall fermion/truncated DW Overlap fermion parameters
  !
    real(DP), private ::        mass = 1.0_DP
    real(DP), private ::   dw_height = 1.0_DP
    real(DP), private ::      ev_max = 10.0e0_DP   ! kernel operator max eigenvalue
    real(DP), private ::      ev_min = 1.0e0_DP    ! kernel operator shifted min eigenvalue
    real(DP), private :: ev_min_true = 1.0e0_DP    ! kernel operator true min eigenvalue
    real(DP), private ::     zol_tol = 0.0001_DP   ! Zolotarev approx tolerance
    real(DP), private ::          a5 = 1.0e0_DP    ! 5th-lattice spacing

    real(DP), private,  allocatable ::   b(:)      ! B term
    real(DP), private,  allocatable ::   c(:)      ! C term

    real(DP), private,  allocatable ::  KP(:,:), KN(:,:)   ! K = 4 - M + (1-M5)/(B+C*M5), KP=chiral P^-, KN=P^+ part
    real(DP), private,  allocatable :: MAP(:,:),MAN(:,:)   ! MA=(1 - M5)
    real(DP), private,  allocatable :: MBP(:,:),MBN(:,:)   ! MB=(B+C*M5)
    real(DP), private,  allocatable :: INVMBP(:,:),INVMBN(:,:)       ! INVMB  = (B+C*M5)^-1
    real(DP), private,  allocatable :: INVDIAGP(:,:),INVDIAGN(:,:)   ! NVDIAG = K^-1
    real(DP), private,  allocatable :: INV_NORMAL_DIAGP(:,:),INV_NORMAL_DIAGN(:,:)   ! NVDIAG=(K'K+4)^-1

    integer,  private ::        NS = 4           ! extra dimension/flavor extent  [1,2,...,NS]
    integer,  private ::   dw_type = DW_TYPE_SHAMIR
  end type

  interface new
  !
  ! create dwf parameter
  !
    module procedure new_dwq
  end interface

  interface delete
  !
  ! delete dwf paraemter
  !
    module procedure delete_dwq
  end interface

  interface read
  ! 
  ! read dw fermion parameter
  !
    module procedure read_dwq
  end interface

  interface print
  ! 
  ! pirnt dw fermion parameter
  !
    module procedure print_dwq
  end interface

  interface save_config
  ! 
  ! save dw fermion parameter on config file
  !
    module procedure save_config_dwq
  end interface

  interface get_NS
  ! 
  ! get the size of 5th direction
  !
    module procedure get_NS_dwq
  end interface

  interface get_mass
  ! 
  ! get dwf mass parameter
  !
    module procedure get_mass_dwq
  end interface

  interface set_mass
  ! 
  ! set dwf mass parameter
  !
    module procedure set_mass_dwq
  end interface

  interface set_coef
  ! 
  ! set dwf operator parameters b(:) and c(:)
  !
    module procedure set_coef_dwq
  end interface

  interface assign_mult_gamma5
  ! 
  ! assign and multiply DWF-gamma5 operator on dwf field
  !
    module procedure assign_mult_gamma5_dwq
  end interface

  interface accum_mult_gamma5
  ! 
  ! accumulate multiply DWF-gamma5 operator on dwf field.
  !
    module procedure accum_mult_gamma5_dwq
  end interface

  interface assign_mult_dwf
  !
  ! assign and multiply dwf operator on dwf field
  !
    module procedure assign_mult_dwq_all        ! normal version
    module procedure assign_mult_lowprojd_dwq   ! lowmode-shifted/projected version
  end interface

  interface assign_inv_mult_dwf
  !
  ! assign and inverse-multiply dwf operator on dwf field
  !
    module procedure assign_inv_mult_dwq_all        ! normal version
    module procedure assign_inv_mult_lowprojd_dwq   ! lowmode-shifted/projected version
  end interface

  interface assign_mult_impterm_dwf
  !
  ! assign and multiply Chiu-improvement term (B+ C M5/2) on dwf field
  !
    module procedure assign_mult_impterm_dwq
  end interface

contains

subroutine new_dwq_low_modes(this)
!
! create/initialize dwf low-modes container
!
  implicit none
  type(dwf_low_modes), intent(inout) :: this
  integer :: i
  if (this%NEV < 0) then
    call error_stop("Negative number is not valid. (dwf_low_modes).")
  endif
  if (allocated(this%W))  then
    call delete(this%W)
    deallocate(this%W)
  endif
  if (allocated(this%XX)) deallocate(this%XX)

  allocate(this%W)
  call new(this%W,this%NEV)

  allocate(this%XX(1:this%NEV,1:this%NEV))
  this%XX(:,:) = Z0
  return
end subroutine

subroutine delete_dwq_low_modes(this)
!
! delete dwf low-modes container
!
  implicit none
  type(dwf_low_modes), intent(inout) :: this
  integer :: i
  if (allocated(this%W))  then
    call delete(this%W)
    deallocate(this%W)
  endif
  if (allocated(this%XX)) deallocate(this%XX)
  return
end subroutine

subroutine get_dwq_low_modes(this,quark,iout,tol,iiter,u)
!
! compute low-mode vectors for dwf chiral-symmetory improvement
! and eigen-value range
!
  use f95_lapack, only : LA_HESV
  use quark_ovlp_kern_eigmodes_class
  implicit none
  type(dwf_low_modes),     intent(inout) :: this
  class(quark_domainwall), intent(inout) :: quark
  integer,                 intent(in)    :: iout
  integer,                 intent(inout) :: iiter
  real(DP),                intent(in)    :: tol
  type(vfield_gluon_wg),   intent(in)    :: u

  type(quark_ovlp_kern_eigmodes), allocatable :: emodes
  type(field_quark_wg), allocatable :: w,gv
  integer, parameter :: NUM_MAX_EVAL = 5
  integer     :: i,j,NEV,INFO
  real(DP)    :: ETMP(this%NEV),rtmp0,rtmp1,rtmp2
  complex(DP) :: MTMP(this%NEV,this%NEV)
  character(CHARLEN) :: str

  NEV = this%NEV

!
! Compute Lower eigen pairs from the sign kernel operator, construct chiral-symemtry
! improvement vectors for domainwall fermions
!
  allocate(emodes)
  emodes%NEV = NEV
  call new(emodes)
  emodes%which_modes = SIGN_KERNEL_LOW_MODES
  if (is_kernel_type_shamir(quark)) then
    emodes%a5 = quark%a5
  else
    emodes%a5 = 0.0d0
  endif
  emodes%m0 = quark%dw_height

#ifdef _DEBUG
  emodes%debug = .true.
#endif

  call get_sign_kernel_eigen_modes(emodes,iout,tol,iiter,u)

  allocate(w,gv)
  call new(w)
  call new(gv)

  !-------------------------------------
  ! construct lowmode subspace vectors
  ! for domainwall fermion
  !-------------------------------------
  ETMP(:)   = 0.0_DP
  MTMP(:,:) = Z0
  do i=1,NEV
    call assign_mult_gamma5(gv,emodes%V(i))   ! gv = gamma5 V(i)

   
    if (is_kernel_type_shamir(quark)) then

      !
      ! w_i = (a5/2)*(Dw + 2/a5) gamma_5 v_i
      !
      rtmp0 = quark%dw_height - 2.0_DP/quark%a5
      call assign_mult_wd(rtmp0,w,gv,u)       ! w := (Dw(m0) + 2/a5) gamma_5 v(i)
      call accum_mult(w,quark%a5/2.0_DP)      ! w := (a5/2) w = (Dw(m0)*a5/2 + 1) gamma_5 v(i)
      call conv_wqf_to_dwqf(w,this%W,i)       ! W(i) := w

      !
      ! X(i,i) = - a * v_i'w_i    (a=a5/2)
      !
      MTMP(i,i) = - (quark%a5/2.0_DP)*prod(emodes%V(i),w)

      !
      ! X(j,i) = - a * v_j'w_i    (a=a5/2)
      !
      do j=i+1,NEV
        MTMP(j,i) = - (quark%a5/2.0_DP)*prod(emodes%V(j),w)
        MTMP(i,j) = conjg(MTMP(j,i))
      enddo

    else

      !
      ! W(i) := gamma_5 v_i
      !
      call conv_wqf_to_dwqf(gv,this%W,i)

    endif

  enddo

  !-------------------------
  ! get miniumum eigenvalue
  !-------------------------
  do i=1,NEV
    ETMP(i) = ABS(emodes%EV(i))
  enddo
  rtmp0 = MINVAL(ETMP(:))
  quark%ev_min_true = rtmp0

  quark%ev_min = MAXVAL(ETMP(:))

  if (0 == nodeid) then
    write(iout,'("%EVALMIN_TRUE",E24.15)') quark%ev_min_true
    write(iout,'("%EVALMIN     ",E24.15)') quark%ev_min
  endif

  !--------------------
  ! X = 1/S - a*V'W
  !--------------------
  do i=1,nev
    !
    ! shift low-modes
    !
    rtmp0 = emodes%EV(i)
    rtmp1 = 2*SIGN(quark%ev_min,rtmp0) - rtmp0      ! 2*|lambda_max|*sign(lambda_i) - lambda_i

    MTMP(i,i) = MTMP(i,i) + 1.0_DP/rtmp1
    if (0 == nodeid) &
 &    write(iout,'("%EVAL",I3,E24.15," => EVAL_SHFT=",E24.15)')i,rtmp0,rtmp1
  enddo

  do j=1,NEV
  do i=1,NEV
    if (0 == nodeid) &
 &    write(iout,'("%MAT",2I3,2E24.15)')i,j,MTMP(i,j)
  enddo
  enddo
  this%XX(:,:) = Z0
  do i=1,NEV
    this%XX(i,i) = Z1
  enddo
  call LA_HESV(MTMP,this%XX,INFO=INFO)   ! LV%XX = inv(X)
  if (0 /= INFO) then
    write(str,'("LA_HESV error (@get_dwq_low_mode). INFO=",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(TRIM(str))
  endif
  do j=1,NEV
  do i=1,NEV
    if (0 == nodeid) &
 &    write(iout,'("%MATINV",2I3,2E24.15)')i,j,this%XX(i,j)
  enddo
  enddo

!
! Compute High modes.  
!
! The Largest eigen value is requred to estimate the eigenvalue range of the sign kerenel operator
! and The eigen vectors are dropped.
!
  call delete(emodes)
  deallocate(w,gv)

  emodes%NEV = this%NEV
  call new(emodes)
  emodes%which_modes = SIGN_KERNEL_HIGH_MODES
  if (is_kernel_type_shamir(quark)) then
    emodes%a5 = quark%a5
  else
    emodes%a5 = 0.0d0
  endif
  emodes%m0 = quark%dw_height
#ifdef _DEBUG
  emodes%debug = .true.
#endif
  call get_sign_kernel_eigen_modes(emodes,iout,tol,iiter,u)

  do i=1,emodes%NEV
    ETMP(i) = ABS(emodes%EV(i))
  enddo
  quark%ev_max = MAXVAL(ETMP(1:emodes%NEV))

  if (0 == nodeid) then
    write(iout,'("%EVALMAX     ",E24.15)') quark%ev_max
    do i=1,emodes%NEV
      rtmp0 = emodes%EV(i)
      write(iout,'("%EVAL",I3,E24.15)')i,rtmp0
    enddo
  endif
  
  call delete(emodes)
  deallocate(emodes)
  
  return
end subroutine

subroutine assign_mult_gamma5_dwq(gy,y)
!
! Multiply Dowmainwall fermion GAMMA5
!
!  gy <= Gamma5 y
!
! this includes permutaion in extra-dimension.
!
!\[
! gy(n,s) <= \gamma_5 y(n,NS-s+1)
!\]
!
! with $ s=1,\cdots,NS. $ 
!
  implicit none
  type(field_dw_quark_wg), intent(inout) :: gy
  type(field_dw_quark_wg), intent(in)    ::  y
  integer :: iw,ix,iy,iz,it,NS,ic,jw
  complex(DP) :: a1,a2,a3,a4
  if ( (.not.is_allocated(gy)) .or. &
 &     (.not.is_allocated(y))  .or. &
 &     (.not.is_same_size(gy,y))  ) then
    call error_stop("5-dim size of gy,y is wrong in assign_mult_gamma5_dwq.")
  endif
  NS = y%NS
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,jw,ic,a1,a2,a3,a4)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT
  do iw=1,NS
    jw = NS-iw+1
    do ic=1,COL
      a1 = y%s(iw,it,iz,iy,ix)%y(ic,1)
      a2 = y%s(iw,it,iz,iy,ix)%y(ic,2)
      a3 = y%s(iw,it,iz,iy,ix)%y(ic,3)
      a4 = y%s(iw,it,iz,iy,ix)%y(ic,4)
      gy%s(jw,it,iz,iy,ix)%y(ic,1) = a3
      gy%s(jw,it,iz,iy,ix)%y(ic,2) = a4
      gy%s(jw,it,iz,iy,ix)%y(ic,3) = a1
      gy%s(jw,it,iz,iy,ix)%y(ic,4) = a2
    enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  return
end subroutine

subroutine accum_mult_gamma5_dwq(y)
!
! Multiply Dowmainwall fermion GAMMA5
!
! y <= Gamma5 y
!
! this includes permutaion in extra-dimension.
!
!\[
! y(n,s) <= \gamma_5 y(n,NS-s+1)
!\]
!
! with $ s=1,\cdots,NS. $ 
!
  implicit none
  type(field_dw_quark_wg), intent(inout) :: y
  integer :: iw,ix,iy,iz,it,NS,ic,jw
  complex(DP) :: a1,a2,a3,a4
  complex(DP) :: b1,b2,b3,b4
  if ( (.not.is_allocated(y)) ) then
    call error_stop("5-dim size of y is wrong in accum_mult_gamma5_dwq.")
  endif
  NS = y%NS
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,jw,ic,a1,a2,a3,a4,b1,b2,b3,b4)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT
    do iw=1,NS/2
      jw = NS-iw+1
      do ic=1,COL
        a1 = y%s(iw,it,iz,iy,ix)%y(ic,1)
        a2 = y%s(iw,it,iz,iy,ix)%y(ic,2)
        a3 = y%s(iw,it,iz,iy,ix)%y(ic,3)
        a4 = y%s(iw,it,iz,iy,ix)%y(ic,4)

        b1 = y%s(jw,it,iz,iy,ix)%y(ic,1)
        b2 = y%s(jw,it,iz,iy,ix)%y(ic,2)
        b3 = y%s(jw,it,iz,iy,ix)%y(ic,3)
        b4 = y%s(jw,it,iz,iy,ix)%y(ic,4)

        y%s(jw,it,iz,iy,ix)%y(ic,1) = a3
        y%s(jw,it,iz,iy,ix)%y(ic,2) = a4
        y%s(jw,it,iz,iy,ix)%y(ic,3) = a1
        y%s(jw,it,iz,iy,ix)%y(ic,4) = a2

        y%s(iw,it,iz,iy,ix)%y(ic,1) = b3
        y%s(iw,it,iz,iy,ix)%y(ic,2) = b4
        y%s(iw,it,iz,iy,ix)%y(ic,3) = b1
        y%s(iw,it,iz,iy,ix)%y(ic,4) = b2
      enddo
    enddo
    if (mod(NS,2)==1) then
      iw = NS/2 + 1
      do ic=1,COL
        a1 = y%s(iw,it,iz,iy,ix)%y(ic,1)
        a2 = y%s(iw,it,iz,iy,ix)%y(ic,2)
        a3 = y%s(iw,it,iz,iy,ix)%y(ic,3)
        a4 = y%s(iw,it,iz,iy,ix)%y(ic,4)
        y%s(iw,it,iz,iy,ix)%y(ic,1) = a3
        y%s(iw,it,iz,iy,ix)%y(ic,2) = a4
        y%s(iw,it,iz,iy,ix)%y(ic,3) = a1
        y%s(iw,it,iz,iy,ix)%y(ic,4) = a2
      enddo
    endif
  enddo
  enddo
  enddo
  enddo
  return
end subroutine

subroutine assign_proj_4dto5d(py,y)
!
! Multiply permutation + chiral projection
!
!-------------------------------------------------------
! Normal permutation + chiral projection (4D->5D)
!
!* $ py(1)      = (P^{-} y)/2 $ 
!* $ py(2:NS-1) = 0  $  
!* $ py(NS)     = (P^{+} y)/2 $  
!
!-------------------------------------------------------
!
!* $  (1-\gamma_5) y = P^{-} y $ .
!* $  (1+\gamma_5) y = P^{+} y $ .
!
  implicit none
  type(field_dw_quark_wg), intent(inout) :: py
  type(field_quark_wg),    intent(in)    ::  y
  integer :: ix,iy,iz,it,ic,iw,NS,itb,ieo
  complex(DP) :: a1,a2,a3,a4,w1,w2,v1,v2
  if ( (.not.is_allocated(py)) ) then
    call error_stop("5-dim size of py is wrong in assign_proj_4dto5d.")
  endif
  NS = py%NS
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,itb,ieo,iw,ic, &
!$OMP         a1,a2,a3,a4,w1,w2,v1,v2)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do it=1,NT
    ieo = mod(ipeo+it+iz+iy+ix,2)
    itb = it/2
    do ic=1,COL
      a1 = y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,1)
      a2 = y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,2)
      a3 = y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,3)
      a4 = y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,4)
      w1 = (a1 - a3)/2.0_DP
      w2 = (a2 - a4)/2.0_DP
      py%s( 1,it,iz,iy,ix)%y(ic,1) =  w1
      py%s( 1,it,iz,iy,ix)%y(ic,2) =  w2
      py%s( 1,it,iz,iy,ix)%y(ic,3) = -w1
      py%s( 1,it,iz,iy,ix)%y(ic,4) = -w2

      v1 = (a1 + a3)/2.0_DP
      v2 = (a2 + a4)/2.0_DP
      py%s(NS,it,iz,iy,ix)%y(ic,1) = v1
      py%s(NS,it,iz,iy,ix)%y(ic,2) = v2
      py%s(NS,it,iz,iy,ix)%y(ic,3) = v1
      py%s(NS,it,iz,iy,ix)%y(ic,4) = v2
    enddo
    do iw=2,NS-1
      py%s(iw,it,iz,iy,ix)%y(:,:) = Z0
    enddo
  enddo
  enddo
  enddo
  enddo
  return
end subroutine

subroutine assign_proj_5dto4d(y,py)
!
! Multiply permutation + chiral projection
!
!-------------------------------------------------------
! Normal permutation + chiral projection (5D->4D)
!
!* $  y  = (P^{-} py(1) + P^{+} py(NS))/2  $  
!
!-------------------------------------------------------
!
!* $  (1-\gamma_5) y = P^{-} y $ .
!* $  (1+\gamma_5) y = P^{+} y $ .
!
  implicit none
  type(field_quark_wg),    intent(inout) ::  y
  type(field_dw_quark_wg), intent(in)    :: py
  integer :: ix,iy,iz,it,ieo,itb,ic,iw,NS
  complex(DP) :: a1,a2,a3,a4,w1,w2,v1,v2
  if ( (.not.is_allocated(py)) ) then
    call error_stop("5-dim size of py is wrong in assign_proj_5dto4d.")
  endif
  call new(y)
  NS = py%NS
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,ieo,itb,ic, &
!$OMP         a1,a2,a3,a4,w1,w2,v1,v2)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do it=1,NT
    ieo = mod(ipeo+it+iz+iy+ix,2)
    itb = it/2
    do ic=1,COL
      a1 = py%s( 1,it,iz,iy,ix)%y(ic,1)
      a2 = py%s( 1,it,iz,iy,ix)%y(ic,2)
      a3 = py%s( 1,it,iz,iy,ix)%y(ic,3)
      a4 = py%s( 1,it,iz,iy,ix)%y(ic,4)
      w1 = (a1 - a3)/2.0_DP
      w2 = (a2 - a4)/2.0_DP
      a1 = py%s(NS,it,iz,iy,ix)%y(ic,1)
      a2 = py%s(NS,it,iz,iy,ix)%y(ic,2)
      a3 = py%s(NS,it,iz,iy,ix)%y(ic,3)
      a4 = py%s(NS,it,iz,iy,ix)%y(ic,4)
      v1 = (a1 + a3)/2.0_DP
      v2 = (a2 + a4)/2.0_DP
      y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,1) =  w1 + v1
      y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,2) =  w2 + v2
      y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,3) = -w1 + v1
      y%eo(ieo)%s(itb,iz,iy,ix)%y(ic,4) = -w2 + v2
    enddo
  enddo
  enddo
  enddo
  enddo
  return
end subroutine

!--------------------------------------------------------------------
! Domainwall type bound methods
!--------------------------------------------------------------------
function get_NS_dwq(this) result(NS)
!
! return size of extra-dimensiion (extra-flavors)
!
  implicit none
  class(quark_domainwall), intent(in) :: this
  integer :: NS
  NS = this%NS
  return
end function

function get_mass_dwq(this) result(mass)
!
! return quark mass
!
  implicit none
  class(quark_domainwall), intent(in) :: this
  real(DP) :: mass
  mass = this%mass
  return
end function

subroutine set_mass_dwq(this,mass)
!
! set quark mass
!
  implicit none
  class(quark_domainwall), intent(inout) :: this
  real(DP), intent(in) :: mass
  this%mass = mass
  return
end subroutine

function is_type_zolotarev(this) result(flag)
!
! query sign approximation type, Zolotarev or not
!
  implicit none
  class(quark_domainwall), intent(in) :: this
  logical :: flag
  if (this%dw_type == DW_TYPE_ZOLOTAREV_SHAMIR .or. this%dw_type == DW_TYPE_ZOLOTAREV_CHIU) then
    flag = .true.
  else
    flag = .false.
  endif
  return
end function

function is_kernel_type_shamir(this) result(flag)
!
! query sign kernel operator type, Shamir or not
!
  implicit none
  class(quark_domainwall), intent(in) :: this
  logical :: flag
  if (this%dw_type == DW_TYPE_ZOLOTAREV_SHAMIR .or. this%dw_type == DW_TYPE_SHAMIR) then
    flag = .true.
  else
    flag = .false.
  endif
  return
end function

subroutine new_dwq(this,id)
!
! initialize dwf parameter
!
  implicit none
  class(quark_domainwall),  intent(inout) :: this
  integer,                  intent(in) :: id

  this%dw_height = 1.0d0
  this%mass      = 0.1d0
  this%NS        = 4

  call new(this%quark_wilson,id)
  this%action_name = ACTION_NAME

  call delete_mass_term_dwq(this)

  return
end subroutine

subroutine delete_dwq(this)
!
! delete dwf parameter
!
  implicit none
  class(quark_domainwall), intent(inout) :: this

  call delete_mass_term_dwq(this)

  return
end subroutine

subroutine allocate_mass_term_dwq(this)
!
! allocate 5D-mass term arrays and parameters for dwf operator
!
  implicit none
  class(quark_domainwall), intent(inout) :: this
  if (.not.allocated(this%b)) allocate(this%b(this%NS))
  if (.not.allocated(this%c)) allocate(this%c(this%NS))

  if (.not.allocated(this%KP))               allocate(this%KP(this%NS,this%NS))
  if (.not.allocated(this%KN))               allocate(this%KN(this%NS,this%NS))
  if (.not.allocated(this%MAP))              allocate(this%MAP(this%NS,this%NS))
  if (.not.allocated(this%MAN))              allocate(this%MAN(this%NS,this%NS))
  if (.not.allocated(this%MBP))              allocate(this%MBP(this%NS,this%NS))
  if (.not.allocated(this%MBN))              allocate(this%MBN(this%NS,this%NS))
  if (.not.allocated(this%INVMBP))           allocate(this%INVMBP(this%NS,this%NS))
  if (.not.allocated(this%INVMBN))           allocate(this%INVMBN(this%NS,this%NS))
  if (.not.allocated(this%INVDIAGP))         allocate(this%INVDIAGP(this%NS,this%NS))
  if (.not.allocated(this%INVDIAGN))         allocate(this%INVDIAGN(this%NS,this%NS))
  if (.not.allocated(this%INV_NORMAL_DIAGP)) allocate(this%INV_NORMAL_DIAGP(this%NS,this%NS))
  if (.not.allocated(this%INV_NORMAL_DIAGN)) allocate(this%INV_NORMAL_DIAGN(this%NS,this%NS))

  return
end subroutine

subroutine delete_mass_term_dwq(this)
!
! delete dwf 5D-mass term arrays and parameters
!
  implicit none
  class(quark_domainwall), intent(inout) :: this

  if (allocated(this%b))  deallocate(this%b)
  if (allocated(this%c))  deallocate(this%c)

  if (allocated(this%KP))               deallocate(this%KP)
  if (allocated(this%KN))               deallocate(this%KN)
  if (allocated(this%MAP))              deallocate(this%MAP)
  if (allocated(this%MAN))              deallocate(this%MAN)
  if (allocated(this%MBP))              deallocate(this%MBP)
  if (allocated(this%MBN))              deallocate(this%MBN)
  if (allocated(this%INVMBP))           deallocate(this%INVMBP)
  if (allocated(this%INVMBN))           deallocate(this%INVMBN)
  if (allocated(this%INVDIAGP))         deallocate(this%INVDIAGP)
  if (allocated(this%INVDIAGN))         deallocate(this%INVDIAGN)
  if (allocated(this%INV_NORMAL_DIAGP)) deallocate(this%INV_NORMAL_DIAGP)
  if (allocated(this%INV_NORMAL_DIAGN)) deallocate(this%INV_NORMAL_DIAGN)

  return
end subroutine

subroutine read_dwq(this,iout)
!
! read dwf paramters
!
  implicit none
  class(quark_domainwall), intent(inout) :: this
  integer, intent(in) :: iout
  character(CHARLEN) :: str
  integer :: nflavor
  if (nodeid==0) then
    read(iout,*) nflavor
    read(iout,*) this%mass
    read(iout,*) this%dw_height
    read(iout,*) this%NS
    read(iout,*) this%dw_type
    read(iout,*) this%zol_tol
  endif

#ifndef _singlePU
  call comlib_bcast(nflavor,0)
  call comlib_bcast(this%mass,0)
  call comlib_bcast(this%dw_height,0)
  call comlib_bcast(this%NS,0)
  call comlib_bcast(this%dw_type,0)
  call comlib_bcast(this%zol_tol,0)
#endif

  if (nflavor < 0 .or. mod(nflavor,2) /= 0 ) then
    write(str,'("nflavor is invalid. nf=",I4,A,I5)')nflavor,__FILE__,__LINE__
    call error_stop(str)
  endif
  if (this%mass < 0.0_DP) then
    write(str,'("mass is invalid. mass=",E24.15,A,I5)')this%mass,__FILE__,__LINE__
    call error_stop(str)
  endif
  if (this%dw_height < 0.0_DP) then
    write(str,'("dw_height is invalid. dw_height=",E24.15,A,I5)')this%dw_height,__FILE__,__LINE__
    call error_stop(str)
  endif
  if (this%NS < 0.0_DP) then
    write(str,'("NS is invalid. NS=",I5,A,I5)')this%NS,__FILE__,__LINE__
    call error_stop(str)
  endif
  if (this%dw_type < 0.0_DP) then
    write(str,'("dw_type is invalid. dw_type=",I5,A,I5)')this%dw_type,__FILE__,__LINE__
    call error_stop(str)
  endif
  if (this%zol_tol < 0.0_DP) then
    write(str,'("zol_tol is invalid. zol_tol=",E24.5,A,I5)')this%zol_tol,__FILE__,__LINE__
    call error_stop(str)
  endif
  call set_nflavor(this%quark_wilson,nflavor)

  return
end subroutine
  
subroutine print_dwq(this)
!
! print out dwf paramters
!
  implicit none
  class(quark_domainwall), intent(in) :: this
  integer :: nflavor
  character(CHARLEN) :: str
  nflavor = get_nflavor(this%quark_wilson)
  if (nflavor /= 2) then
    write(str,'("Num of Flavor should be 2:",I3)')nflavor
    call error_stop(TRIM(str))
  endif
  if (nodeid==0) then
  write(*,'(16X,"       Num Flavor :",I4)')   nflavor
  write(*,'(16X,"             Mass :",F12.8)')this%mass
  write(*,'(16X,"Domainwall Height :",F12.8)')this%dw_height
  write(*,'(16X,"               NS :",I4)')   this%NS
  write(*,'(16X,"               a5 :",F12.8)')this%a5
  write(*,'(16X,"             Type : ",A)')TRIM(DW_TYPE_NAME(this%dw_type))
  if (is_type_zolotarev(this)) then
  write(*,'(12X,"Zolotarev approx. tol :",E12.4)')this%zol_tol
  endif
  endif

  return
end subroutine

subroutine save_config_dwq(this,iout)
!
! save dwf paramters on config file
!
  implicit none
  class(quark_domainwall), intent(in) :: this
  integer, intent(in) :: iout
  integer :: nflavor
  nflavor = get_nflavor(this%quark_wilson)
  if (nodeid==0) then
  write(iout)nflavor
  write(iout)this%mass
  write(iout)this%dw_height
  write(iout)this%NS
  endif
  return
end subroutine

subroutine set_coef_dwq(this)
!
! set dwf chiral-symmetry improvemnt parameters
!
  use signfunc_mod
  implicit none
  class(quark_domainwall), intent(inout) :: this
  real(DP) :: eps
  type(zolotarev_signfunc) :: sf
  real(DP), parameter :: LMAX=1.01_DP, LMIN=0.99_DP
  integer:: iw

  eps = this%zol_tol

  call delete_mass_term_dwq(this)

  select case(this%dw_type)
  case(DW_TYPE_SHAMIR)

    call allocate_mass_term_dwq(this)

    this%b(:) = this%a5
    this%c(:) = 0.0_DP

  case(DW_TYPE_BORICI)

    call allocate_mass_term_dwq(this)

    this%b(:) = this%a5
    this%c(:) = this%a5

  case(DW_TYPE_ZOLOTAREV_SHAMIR)

    call new(sf,this%ev_max*LMAX,this%ev_min*LMIN,eps)
    this%NS = sf%NS

    call allocate_mass_term_dwq(this)

    this%b(:) = (2*sf%omg(:)/(this%ev_min)+this%a5)/2.0_DP
    this%c(:) = (2*sf%omg(:)/(this%ev_min)-this%a5)/2.0_DP

    if (0==nodeid) then
      write(*,'(80("-"))')
      write(*,'(" ZOLOTAREV_SHAMIR")')
      write(*,'("         EV max:",E24.15)')this%ev_max
      write(*,'("         EV min:",E24.15)')this%ev_min
      write(*,'("    EV max*LMAX:",E24.15)')this%ev_max*LMAX
      write(*,'("    EV min*LMIN:",E24.15)')this%ev_min*LMIN
      write(*,'("           EPS :",E24.15)')eps
      write(*,'("           ERR :",E24.15)')sf%delta
      write(*,'(" reallocated NS:",I3)')this%NS
      do iw=1,this%NS
      write(*,'("       omg(",I3,"):",E24.15," b(",I3,"):",E24.15," c(",I3,"):",E24.15)') &
  &       iw,sf%omg(iw),iw,this%b(iw),iw,this%c(iw)
      enddo
      write(*,'(80("-"))')
    endif
    call delete(sf)

  case(DW_TYPE_ZOLOTAREV_CHIU)

    call new(sf,this%ev_max*LMAX,this%ev_min*LMIN,eps)
    this%NS = sf%NS

    call allocate_mass_term_dwq(this)

    this%b(:) = sf%omg(:)/(this%ev_min)
    this%c(:) = sf%omg(:)/(this%ev_min)

    if (0==nodeid) then
      write(*,'(80("-"))')
      write(*,'(" ZOLOTAREV_CHIU")')
      write(*,'("         EV max:",E24.15)')this%ev_max
      write(*,'("         EV min:",E24.15)')this%ev_min
      write(*,'("    EV max*LMAX:",E24.15)')this%ev_max*LMAX
      write(*,'("    EV min*LMIN:",E24.15)')this%ev_min*LMIN
      write(*,'("           EPS :",E24.15)')eps
      write(*,'("           ERR :",E24.15)')sf%delta
      write(*,'(" reallocated NS:",I3)')this%NS
      do iw=1,this%NS
      write(*,'("       omg(",I3,"):",E24.15," b(",I3,"):",E24.15," c(",I3,"):",E24.15)') &
  &       iw,sf%omg(iw),iw,this%b(iw),iw,this%c(iw)
      enddo
      write(*,'(80("-"))')
    endif
    call delete(sf)

  end select

  call make_mass_term(this)

  return
end subroutine

subroutine assign_mult_impterm_dwq(this,dq,q,dagger)
! 
! Assign and multiply improvement term
!
!  dq = (B + C M5) q
!
! or
!
!  dq = (B + M5^dag C) q
!
!  B^dag = B, C^dag = C
!
!  M5: hopping in 5th direction
!
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  logical, optional, intent(in) :: dagger
  type(su3fv_spinor) :: yy,yt(this%NS)
  complex(DP) :: ctmp(COL,SPIN)
  integer :: ic
  integer :: NS,ix,iy,iz,it,iw
  integer :: kw
  logical :: is_dagger

  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif

  NS = q%NS

  if (DW_TYPE_SHAMIR /= this%dw_type) then

    if (.not.is_dagger) then

      !
      ! dq = (B + C M5) q = MB q
      !
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,kw,yt,yy)
      do ix=1,NX
      do iy=1,NY         
      do iz=1,NZ
      do it=1,NT

        do iw=1,NS
          yt(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
          yt(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
        enddo
        do iw=1,NS
          yy%y(:,:) = Z0
          do kw=1,NS
            yy%y(:,1:2) = yy%y(:,1:2)+this%MBP(iw,kw) * yt(kw)%y(:,1:2)
            yy%y(:,3:4) = yy%y(:,3:4)+this%MBN(iw,kw) * yt(kw)%y(:,3:4)
          enddo
          dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yy%y(:,1:2) + yy%y(:,3:4))*0.5_DP 
          dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yy%y(:,1:2) + yy%y(:,3:4))*0.5_DP 
        enddo

      enddo
      enddo
      enddo
      enddo


    else

      !
      ! dq = (B + M5^dag C) q = MB^dag q
      !
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,kw,yt,yy)
      do ix=1,NX
      do iy=1,NY         
      do iz=1,NZ
      do it=1,NT

        do iw=1,NS
          yt(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
          yt(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
        enddo
        do iw=1,NS
          yy%y(:,:) = Z0
          do kw=1,NS
            yy%y(:,1:2) = yy%y(:,1:2)+this%MBP(kw,iw) * yt(kw)%y(:,1:2)
            yy%y(:,3:4) = yy%y(:,3:4)+this%MBN(kw,iw) * yt(kw)%y(:,3:4)
          enddo
          dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yy%y(:,1:2) + yy%y(:,3:4))*0.5_DP 
          dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yy%y(:,1:2) + yy%y(:,3:4))*0.5_DP 
        enddo

      enddo
      enddo
      enddo
      enddo

    endif

  else

    call assign(dq,q)

  endif

  return
end subroutine

subroutine assign_mult_dwq_all(this,dq,q,u,dagger)
!
! assign and multiply dwf operaotr
!
! dq <= Ddw q
!
! or
!
! dq <= Ddw^dag q
!
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_gluon_wg),   intent(in)    ::  u
  logical,       optional, intent(in) :: dagger   ! if present, Ddw^dagger is multiplyed.
  logical :: is_dagger
  if (present(dagger)) then
    is_dagger=.true.
  else
    is_dagger=.false.
  endif
  if (.not.is_dagger) then
    call        assign_mult_dwq(this,dq,q,u)
  else
    call assign_mult_dagger_dwq(this,dq,q,u)
  endif
  return
end subroutine

subroutine assign_mult_dwq(this,dq,q,u)
!
! Assign and multiply Domainwall operator
!
! dq <= Ddw q
!
!\[
!  (D_{DW}(m_q))(n,m)_{s,r} = ((4-M) - \frac{1}{2} M_{hop}(n,m)) (B+C M_5)+(1-M_5)
!\]
!
!\[ M_5 = \frac{1}{2} \left[ (1-\gamma_5) \delta_{s+1,r}
!                          + (1+\gamma_5) \delta_{s-1,r} \right.  \] \[ \left.
!                      - m_q (1-\gamma_5) \delta_{s,N}\delta_{1,r} 
!                      - m_q (1+\gamma_5) \delta_{s,1}\delta_{N,r} \right]
!\]
!
  use counter_class
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_dw_quark_wg), allocatable :: aq,bq
  type(su3fv_spinor) :: yt,yw,yy(this%NS)
  complex(DP) :: w(COL,SPIN/2), y(COL,SPIN/2)
  integer :: NS
  integer :: ix,iy,iz,it,itb,itbup,itbdn,ieoxyz,ieo,ioe
  integer :: iw,iwup,iwdn,kup,kdn
  integer :: kw
  real(DP) :: fup,fdn,mq,dwfac, MM

  NS = this%NS
  if ( (q%NS /= NS) .or. (.not. is_same_size(q,dq)) ) then
    call error_stop("5-dim size of q,dq is wrong in assign_mult_dwq.")
  endif

  allocate(aq,bq)
  call new(aq,NS)
  call new(bq,NS)

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,kw,yt,yy)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT

    do iw=1,NS
      yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
      yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
    enddo
    do iw=1,NS
      yt%y(:,:) = Z0
      do kw=1,NS
        yt%y(:,1:2) = yt%y(:,1:2) + this%MAP(iw,kw) * yy(kw)%y(:,1:2)
        yt%y(:,3:4) = yt%y(:,3:4) + this%MAN(iw,kw) * yy(kw)%y(:,3:4)
      enddo
      aq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
      aq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
    enddo

    do iw=1,NS
      yt%y(:,:) = Z0
      do kw=1,NS
        yt%y(:,1:2) = yt%y(:,1:2) + this%MBP(iw,kw) * yy(kw)%y(:,1:2)
        yt%y(:,3:4) = yt%y(:,3:4) + this%MBN(iw,kw) * yy(kw)%y(:,3:4)
      enddo
      bq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
      bq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
    enddo

  enddo
  enddo
  enddo
  enddo

  dwfac = 4.0_DP - this%dw_height
  call copy_boundary(bq)

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,ieo,ioe,ieoxyz,itb,itbup,itbdn,iw,yt,w,y)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT
    itb = it/2
    ieo = mod(ipeo+it+iz+iy+ix,2)
    ioe = 1-ieo
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    itbup=   +ieoxyz 
    itbdn=-(1-ieoxyz)

    do iw=1,NS
#include "mult_dw_hop4_kernel.h90"
      dq%s(iw,it,iz,iy,ix)%y(:,:) = aq%s(iw,it,iz,iy,ix)%y(:,:)  &
  &                                +bq%s(iw,it,iz,iy,ix)%y(:,:)*dwfac &
  &                                          - 0.5_DP*yt%y(:,:)
    enddo

  enddo
  enddo
  enddo
  enddo
  call inc(mult_iter,NS*2)

  call delete(aq)
  call delete(bq)
  deallocate(aq,bq)

  return
end subroutine

subroutine assign_mult_dagger_dwq(this,dq,q,u)
!
! Assign and multiply Domainwall dagger operator
!
! dq <= Ddw' q
!
  use counter_class
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_dw_quark_wg), allocatable :: bq
  type(su3fv_spinor) :: yt,yy(this%NS)
  complex(DP) :: w(COL,SPIN/2), y(COL,SPIN/2)
  integer :: NS
  integer :: ix,iy,iz,it,itb,itbup,itbdn,ieoxyz,ieo,ioe
  integer :: iw,iwup,iwdn,kup,kdn
  integer :: kw
  logical :: dagger
  real(DP) :: fup,fdn,mq,dwfac,MM

  NS = this%NS
  if ( (q%NS /= NS) .or. (.not. is_same_size(q,dq)) ) then
    call error_stop("5-dim size of q,dq is wrong in assign_mult_dagger_dwq.")
  endif

  allocate(bq)
  call new(bq,NS)

  dwfac = 4.0_DP - this%dw_height
  call copy_boundary(q)

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,ieo,ioe,ieoxyz,itb,itbup,itbdn,iw,yt,w,y)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT
    itb = it/2
    ieo = mod(ipeo+it+iz+iy+ix,2)
    ioe = 1-ieo
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    itbup=   +ieoxyz 
    itbdn=-(1-ieoxyz)
    do iw=1,NS
#include "mult_dw_hop4_dagger_kernel.h90"
      bq%s(iw,it,iz,iy,ix)%y(:,:) = q%s(iw,it,iz,iy,ix)%y(:,:)*dwfac - 0.5_DP*yt%y(:,:)
    enddo
  enddo
  enddo
  enddo
  enddo
  call inc(mult_iter,NS*2)

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,kw,yt,yy)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT

    do iw=1,NS
      yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
      yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
    enddo
    do iw=1,NS
      yt%y(:,:) = Z0
      do kw=1,NS
        yt%y(:,1:2) = yt%y(:,1:2) + this%MAP(kw,iw) * yy(kw)%y(:,1:2)
        yt%y(:,3:4) = yt%y(:,3:4) + this%MAN(kw,iw) * yy(kw)%y(:,3:4)
      enddo
      dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
      dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
    enddo

    do iw=1,NS
      yy(iw)%y(:,1:2) = bq%s(iw,it,iz,iy,ix)%y(:,1:2) - bq%s(iw,it,iz,iy,ix)%y(:,3:4)
      yy(iw)%y(:,3:4) = bq%s(iw,it,iz,iy,ix)%y(:,1:2) + bq%s(iw,it,iz,iy,ix)%y(:,3:4)
    enddo
    do iw=1,NS
      yt%y(:,:) = Z0
      do kw=1,NS
        yt%y(:,1:2) = yt%y(:,1:2) + this%MBP(kw,iw) * yy(kw)%y(:,1:2)
        yt%y(:,3:4) = yt%y(:,3:4) + this%MBN(kw,iw) * yy(kw)%y(:,3:4)
      enddo
      dq%s(iw,it,iz,iy,ix)%y(:,1:2) = dq%s(iw,it,iz,iy,ix)%y(:,1:2) + (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
      dq%s(iw,it,iz,iy,ix)%y(:,3:4) = dq%s(iw,it,iz,iy,ix)%y(:,3:4) + (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
    enddo

  enddo
  enddo
  enddo
  enddo

  call delete(bq)
  deallocate(bq)

  return
end subroutine

subroutine assign_inv_mult_dwq_all(this,iout,tol,iiter,dy,y,u,dagger,cron)
!
! Assign and multiply inverse Domainwall operator
!
! dy <= Ddw \ y
!
! or 
!
! dy <= Ddw^dag \ y
!
! using CGNE solver
!
  use solver_class
  use counter_class
#ifdef _PERFMON_
  use perfmon_class
#endif
  implicit none
  class(quark_domainwall), intent(in) :: this
  integer,  intent(in)    :: iout
  real(DP), intent(in)    :: tol
  integer,  intent(inout) :: iiter
  type(field_dw_quark_wg), intent(inout) :: dy
  type(field_dw_quark_wg), intent(inout) ::  y
  type(vfield_gluon_wg),   intent(in)    ::  u
  logical,           optional, intent(in)    :: dagger
  type(chrolog_alg), optional, intent(inout) :: cron
  type(field_dw_quark_wg),  allocatable :: w,v,yy
  type(vfield_dw_gluon_wg), allocatable :: vu
#if _PREC_ == _PREC_EO_
  integer :: ieo,ioe
  integer, parameter :: SLV_NSITE=COL*SPIN*NTH*NZ*NY*NX
#else
  integer, parameter :: SLV_NSITE=COL*SPIN*NT*NZ*NY*NX
#endif
  character(CHARLEN) :: str
  real(DP) :: etime,rtmp0,rtmp1
  type(timer) :: solver_time
  integer :: NS,istat,mult_count
  type(counter) :: imult
  type(cg_alg), allocatable :: solver
  integer :: i,j,NEV
  complex(DP) :: ctmp
  logical :: is_dagger
  logical :: do_cron
#if _PREC_ == _PREC_4DSSOR_
  real(DP), parameter :: omega  = 1.5_DP
  real(DP), parameter :: omg2   = omega - 2.0_DP
  real(DP), parameter :: omginv = 1.0_DP/omega
#endif
#ifdef _PERFMON_
  real(DP) :: fp0,fp1
  type(flop_counter) :: flop
  call new(flop)
#endif


!  write(*,*)"assign_inv_mult_dwq_all"

  call new(solver_time)
  call tic(solver_time)

  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif

  NS = this%NS
  if (NS /= y%NS) then
    call error_stop("this%NY /= y%NS, input y is wrong in assign_in_mult_dwq.")
  endif
  if (NS /= dy%NS) then
    call new(dy,NS)
  endif

  allocate(w,v)
  call new(w,this%NS)
  call new(v,this%NS)

  imult = mult_iter

  !==============================================
  ! CGNE is used to solve
  !
  ! D x = y.
  !
  ! D^dag D x = D^dag y
  !
  ! or
  !
  ! D^dag x = y.
  !
  ! D D^dag x = D y
  !
  !==============================================

  !
  ! set source for CGNE
  !
  ! w = D^dag y
  !
  ! or
  !
  ! w = D y
  !
#if _PREC_ == _PREC_NON_

  if (.not.is_dagger) then
    call assign_mult_dagger_dwq(this,w,y,u)
  else
    call        assign_mult_dwq(this,w,y,u)
  endif

#elif _PREC_ == _PREC_EO_

  ieo = 0
  ioe = 1-ieo
  allocate(vu)
  call conv_f_dwg_wg(u,vu)
  if (.not.is_dagger) then
    call assign_inv_mult_ee_dwq(this,ioe,w,y)                   ! wo = Doo \ yo
    call     assign_mult_eo_dwq(this,ieo,v,w,vu)                ! ve = Deo wo
    call assign_sub(w,y,v,ieo)                                  ! we = ye - ve
    call assign_inv_mult_ee_dwq(this,ieo,v,w)                   ! ve = Dee \ we
    call assign_mult_eoprec_dwq(this,ieo,w,v,vu,dagger=.true.)  ! we = hat{Dee}^dag \ ve
  else
!    write(*,*)"dagger mode"
    allocate(yy)
    call new(yy,this%NS)
    call assign_mult_invmprec_dwq(this,yy,y,dagger=.true.)      ! yy =  MB \ y
    call assign_inv_mult_ee_dwq(this,ioe,v,yy,dagger=.true.)    ! vo = Doo \ yyo
    call     assign_mult_eo_dwq(this,ieo,w,v,vu,dagger=.true.)  ! we = Deo vo
    call assign_sub(v,yy,w,ieo)                                 ! ve = yye - we
    call assign_mult_eoprec_dwq(this,ieo,w,v,vu)                ! we = hat{Dee} \ ve
  endif

#elif _PREC_ == _PREC_4DSSOR_

  allocate(vu)
  call conv_f_dwg_wg(u,vu)
  if (.not.is_dagger) then
    call assign_mult_dagger_dwq(this,v,y,u) 
    call assign_mult_invmprec_dwq(this,w,v,dagger=.true.)
    call assign_inv_mult2_diag_dwq(this,v,w)
    call assign_inv_mult2_forw_4dssor_dwq(this,omega,w,v,vu,forw=.true.)
  else
    call error_stop("assign_inv_mult_dwf for dagger is not implemented.")
  endif

#endif

  !================================
  ! set source and guess to solver
  !================================
  do_cron = .false.
  if (present(cron)) then
    if (has_history(cron)) then
      do_cron = .true.
    endif
  endif

  if (do_cron) then

    !===================================
    ! start Chronological guess solver
    !===================================
    if (nodeid == 0) then
      write(iout,'("#CRON USED")')
#ifdef _DEBUG
      write(*   ,'("#CRON USED")')
#endif
    endif

#if _PREC_ == _PREC_EO_
    call pack(w,cron%src_vec(:),ieo)
#else
    call pack(w,cron%src_vec(:))
#endif

    do
      call solve(cron)
      select case(get_status(cron))
      case (CHRON_OP_NOP)
        cycle
      case (CHRON_OP_DO_MATVEC)

#if _PREC_ == _PREC_NON_
        call unpack(cron%src_vec,w)
        if (.not.is_dagger) then
          ! 
          ! w <= D^dag D w
          ! 
          call        assign_mult_dwq(this,v,w,u)
          call assign_mult_dagger_dwq(this,w,v,u)
        else
          ! 
          ! w <= D D^dag w
          ! 
          call assign_mult_dagger_dwq(this,v,w,u)
          call        assign_mult_dwq(this,w,v,u)
        endif
        call pack(w,cron%dst_vec)

#elif _PREC_ == _PREC_EO_

        call unpack(cron%src_vec,w,ieo)
        if (.not.is_dagger) then
          ! 
          ! w <= D^dag D w
          ! 
          call assign_mult_eoprec_dwq(this,ieo,v,w,vu)
          call assign_mult_eoprec_dwq(this,ieo,w,v,vu,dagger=.true.)
        else
          ! 
          ! w <= D D^dag w
          ! 
          call assign_mult_eoprec_dwq(this,ieo,v,w,vu,dagger=.true.)
          call assign_mult_eoprec_dwq(this,ieo,w,v,vu)
        endif
        call pack(w,cron%dst_vec,ieo)

#elif _PREC_ == _PREC_4DSSOR_

        call unpack(cron%src_vec,w)
        if (.not.is_dagger) then
          call assign_inv_mult2_forw_4dssor_dwq(this,omega,dy,w,vu,forw=.false.) ! dy = BACK w
          call accum_add_mult(w,dy,omg2)                                         !  w = w + dy * omg2
          call assign_inv_mult2_forw_4dssor_dwq(this,omega, v,w,vu,forw=.true.)  !  v = FORW w
          call accum_add(v,dy)                                                   !  v = v + dy
          call accum_mult(v,omginv)                                              !  v = v * omginv
        endif
        call pack(v,cron%dst_vec)

#endif

      case (CHRON_OP_CONVERGED)
        if (nodeid == 0) then
          mult_count = get_count(mult_iter) - get_count(imult)
          write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)') 0, &
 &                              get_residual_norm(cron),mult_count
#ifdef _DEBUG
          write(*   ,'("#",I5," ERR_=",E24.16," MULT_=",I10)') 0, &
 &                              get_residual_norm(cron),mult_count
#endif
        endif
        exit
      end select
    enddo

    allocate(solver)
    call new(solver,NSIZE=SLV_NSITE*NS,guess=GUESS_USE,mode=MODE_NORMAL    ,max_iter=iiter,tol=tol)

    solver%src_vec(:) = cron%src_vec(:)
    solver%dst_vec(:) = cron%dst_vec(:)

  else

    allocate(solver)
    call new(solver,NSIZE=SLV_NSITE*NS,mode=MODE_NORMAL    ,max_iter=iiter,tol=tol)

#if _PREC_ == _PREC_EO_
    call pack(w,solver%src_vec(:),ieo)
#else
    call pack(w,solver%src_vec(:))
#endif

  endif

  !=======================
  ! start CG Solver
  !=======================

#ifdef _PERFMON_
  call start(flop)
#endif
  do 
    call solve(solver)
    istat = get_status(solver)

    select case(istat)
    case (OP_NOP)
      cycle
    case (OP_PROJECTION)
      cycle
    case (OP_DO_MATVEC)

#if _PREC_ == _PREC_NON_
      call unpack(solver%src_vec,w)
      if (.not.is_dagger) then
        !
        ! w <= D^dag D w
        !
        call        assign_mult_dwq(this,v,w,u)
        call assign_mult_dagger_dwq(this,w,v,u)
      else
        !
        ! w <= D D^dag w
        !
        call assign_mult_dagger_dwq(this,v,w,u)
        call        assign_mult_dwq(this,w,v,u)
      endif
      call pack(w,solver%dst_vec)

#elif _PREC_ == _PREC_EO_

      call unpack(solver%src_vec,w,ieo)
#ifdef _PERFMON_
      call read(flop)
      fp0 = flop%flop
#endif
      if (.not.is_dagger) then
        ! 
        ! w <= D^dag D w
        ! 
        call assign_mult_eoprec_dwq(this,ieo,v,w,vu)
        call assign_mult_eoprec_dwq(this,ieo,w,v,vu,dagger=.true.)
      else
        ! 
        ! w <= D D^dag w
        ! 
        call assign_mult_eoprec_dwq(this,ieo,v,w,vu,dagger=.true.)
        call assign_mult_eoprec_dwq(this,ieo,w,v,vu)
      endif
#ifdef _PERFMON_
      call read(flop)
      fp1 = flop%flop
      if (nodeid == 0) write(*,'("@@",E24.16)')fp1-fp0
#endif
      call pack(w,solver%dst_vec,ieo)

#elif _PREC_ == _PREC_4DSSOR_

      call unpack(solver%src_vec,w)
#ifdef _PERFMON_
      call read(flop)
      fp0 = flop%flop
#endif
      if (.not.is_dagger) then
        call assign_inv_mult2_forw_4dssor_dwq(this,omega,dy,w,vu,forw=.false.) ! dy = BACK w
        call accum_add_mult(w,dy,omg2)                                         !  w = w + dy * omg2
        call assign_inv_mult2_forw_4dssor_dwq(this,omega, v,w,vu,forw=.true.)  !  v = FORW w
        call accum_add(v,dy)                                                   !  v = v + dy
        call accum_mult(v,omginv)                                              !  v = v * omginv
      endif
#ifdef _PERFMON_
      call read(flop)
      fp1 = flop%flop
      if (nodeid == 0) write(*,'("@@",E24.16)')fp1-fp0
#endif
      call pack(v,solver%dst_vec)

#endif

    case (OP_PRINT_STATUS)
#ifdef _DEBUG
#ifdef _PERFMON_
      call read(flop)
      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10," FLOP_=",E24.16," ELAPSE_=",E24.15)') &
 &                        get_current_iteration(solver),&
 &                            get_residual_norm(solver),mult_count,flop%flop,flop%interval
        write(   *,'("#",I5," ERR_=",E24.16," MULT_=",I10," FLOP_=",E24.16," ELAPSE_=",E24.15)') &
 &                        get_current_iteration(solver),&
 &                            get_residual_norm(solver),mult_count,flop%flop,flop%interval
      endif
#else
      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)') &
 &                        get_current_iteration(solver),&
 &                            get_residual_norm(solver),mult_count
        write(   *,'("#",I5," ERR_=",E24.16," MULT_=",I10)') &
 &                        get_current_iteration(solver),&
 &                            get_residual_norm(solver),mult_count
      endif
#endif
#endif
      cycle
    case (OP_CONVERGED)

#ifdef _PERFMON_
      call read(flop)
      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",A," iteration converged.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10," FLOP_=",E24.16," ELAPSE_=",E24.15)',advance='no') &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count,flop%flop,flop%interval
      endif
#else
      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",A," iteration converged.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)',advance='no')     &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count
      endif
#endif
      exit      

    case (OP_MAXITER_REACHED)

      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",A," iteration does not converge.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)',advance='no')     &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count
      endif
      write(str,'("solver did not converge.: ",A,I4)')__FILE__,__LINE__
      call error_stop(str)

    case default

      write(str,'("something is wrong in solver.: ",A,I4)')__FILE__,__LINE__
      call error_stop(str)

    end select
  enddo

#ifdef _PERFMON_
  call stop(flop)
  call delete(flop)
#endif

  if (present(cron)) then
    cron%store_vec(:) = solver%dst_vec(:)
    call store(cron)
  endif

  !
  ! construct solution
  !
#if _PREC_ == _PREC_NON_

  call unpack(solver%dst_vec,dy)

#elif _PREC_ == _PREC_EO_

  call unpack(solver%dst_vec,w,ieo)
  if (.not.is_dagger) then
    call     assign_mult_eo_dwq(this,ioe,v,w,vu)     !  vo = Doe we
    call assign_sub(dy,y,v,ioe)                      ! dyo = yo - vo
    call assign_inv_mult_ee_dwq(this,ioe,w,dy)       !  wo = Doo \ dyo
    call assign_mult_invmprec_dwq(this,dy,w)         !  dy = MB \ w
  else
    call assign_inv_mult_ee_dwq(this,ieo,dy,w,dagger=.true.)     ! dye = Dee \ we
    call     assign_mult_eo_dwq(this,ioe,v,dy,vu,dagger=.true.)  !  vo = Doe dye
    call assign_sub(w,yy,v,ioe)                                  !  wo = yyo - vo
    call assign_inv_mult_ee_dwq(this,ioe,dy,w,dagger=.true.)     ! dyo = Doo \ wo
    call delete(yy)
    deallocate(yy)
  endif
  deallocate(vu)

#elif _PREC_ == _PREC_4DSSOR_

  call unpack(solver%dst_vec,w)
  if (.not.is_dagger) then
    call assign_inv_mult2_forw_4dssor_dwq(this,omega,v,w,vu,forw=.false.)
    call assign_mult_invmprec_dwq(this,dy,v)         !  dy = MB \ w
  endif
  deallocate(vu)

#endif

  call delete(v)
  call delete(w)
  deallocate(w,v)

  iiter = get_current_iteration(solver)

  call delete(solver)
  deallocate(solver)

  call toc(solver_time)

  !============================================
  ! NOTE: get_elapse contains comlib_sumcast.
  ! This means that get_elapse should not used
  ! in if-(nodeid==0)-then-endif block.
  !============================================
  etime = get_elapse(solver_time)  
  if (nodeid == 0) then
    write(iout,'(" ETIME_=",E24.16)')etime
  endif

#ifdef _DEBUG
  allocate(w)
  call new(w,this%NS)
  if (.not.is_dagger) then
    call assign_mult_dwq_all(this,w,dy,u)
  else
    call assign_mult_dwq_all(this,w,dy,u,dagger=dagger)
  endif
  call accum_sub(w,y)
  rtmp0 = abs2(w)
  rtmp1 = abs2(y)
  if (nodeid==0) then
    write(*   ,'("%",I5," ERR_=",3E24.16)',advance='no')    &
 &           iiter,sqrt(rtmp0/rtmp1),sqrt(rtmp0),sqrt(rtmp1)
    write(*   ,'(" ETIME_=",E24.16)')etime
    write(iout,'("%",I5," ERR_=",3E24.16)',advance='no')    &
 &           iiter,sqrt(rtmp0/rtmp1),sqrt(rtmp0),sqrt(rtmp1)
    write(iout,'(" ETIME_=",E24.16)')etime
  endif
  call delete(w)
  deallocate(w)
#endif

  return
end subroutine

subroutine make_mass_term(this)
!
! compute 5D massterm from dwf chiral-symmetry improvement parameters
!
  use f95_lapack, only : LA_GESV
  implicit none
  class(quark_domainwall), intent(inout) :: this
  real(DP) :: mq
  real(DP) :: M5P,M5N
  real(DP) :: TMPP(this%NS,this%NS)
  real(DP) :: TMPN(this%NS,this%NS)
  integer  :: NS,iw,jw,kw,is,js,INFO
  character(CHARLEN) :: str

  NS = this%NS
  mq = this%mass

!$OMP PARALLEL DO PRIVATE(jw,iw,M5P,M5N)
  do jw=1,NS
  do iw=1,NS
      M5P = delta(iw+1,jw) - mq*delta(iw,NS)*delta( 1,jw)
      M5N = delta(iw,jw+1) - mq*delta(iw, 1)*delta(NS,jw)
      this%MAP(iw,jw) =            delta(iw,jw) -            M5P
      this%MAN(iw,jw) =            delta(iw,jw) -            M5N
      this%MBP(iw,jw) = this%b(iw)*delta(iw,jw) + this%c(iw)*M5P
      this%MBN(iw,jw) = this%b(iw)*delta(iw,jw) + this%c(iw)*M5N
      this%INVMBP(iw,jw) = 0.0_DP
      this%INVMBN(iw,jw) = 0.0_DP
  enddo
  enddo

!$OMP PARALLEL DO PRIVATE(iw)
  do iw=1,NS
    this%INVMBP(iw,iw) = 1.0_DP
    this%INVMBN(iw,iw) = 1.0_DP
  enddo

!$OMP PARALLEL DO PRIVATE(jw,iw)
  do jw=1,NS
  do iw=1,NS
    TMPP(iw,jw) = this%MBP(iw,jw)
    TMPN(iw,jw) = this%MBN(iw,jw)
  enddo
  enddo
  call LA_GESV(TMPP,this%INVMBP,INFO=INFO)
  if (INFO /= 0) then
    write(str,'("Error stop at LA_GESV for INVMBP: INFO =",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(str)
  endif
  call LA_GESV(TMPN,this%INVMBN,INFO=INFO)
  if (INFO /= 0) then
    write(str,'("Error stop at LA_GESV for INVMBN: INFO =",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(str)
  endif

!
! MATMUL KP = MAP * INVMBP
! MATMUL KN = MAN * INVMBN
!
!$OMP PARALLEL DO PRIVATE(jw,iw,kw)
  do jw=1,NS
  do iw=1,NS
    this%KP(iw,jw) = (4.0_DP-this%dw_height)*delta(iw,jw)
    this%KN(iw,jw) = (4.0_DP-this%dw_height)*delta(iw,jw)
    do kw=1,NS
      this%KP(iw,jw) = this%KP(iw,jw) + this%MAP(iw,kw)*this%INVMBP(kw,jw)
      this%KN(iw,jw) = this%KN(iw,jw) + this%MAN(iw,kw)*this%INVMBN(kw,jw)
    enddo
  enddo
  enddo

!$OMP PARALLEL DO PRIVATE(jw,iw)
  do iw=1,NS
    do jw=1,NS
      this%INVDIAGP(iw,jw) = 0.0_DP
      this%INVDIAGN(iw,jw) = 0.0_DP
    enddo
    this%INVDIAGP(iw,iw) = 1.0_DP
    this%INVDIAGN(iw,iw) = 1.0_DP
  enddo
!$OMP PARALLEL DO PRIVATE(jw,iw)
  do jw=1,NS
  do iw=1,NS
    TMPP(iw,jw) = this%KP(iw,jw)
    TMPN(iw,jw) = this%KN(iw,jw)
  enddo
  enddo
  call LA_GESV(TMPP,this%INVDIAGP,INFO=INFO)
  if (INFO /= 0) then
    write(str,'("Error stop at LA_GESV for INVDIAGP: INFO =",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(str)
  endif
  call LA_GESV(TMPN,this%INVDIAGN,INFO=INFO)
  if (INFO /= 0) then
    write(str,'("Error stop at LA_GESV for INVDIAGN: INFO =",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(str)
  endif

!==================================================
! Diagnal part of Normal equation Ddw'Ddw
!
!  Ddw' Ddw = (K + Hop/2)' (K + Hop/2)
!           = K'K + K'Hop/2 + Hop'K/2 + Hop'Hop/4
!
! Diag(Ddw'Ddw) = K'K + 4
!
!==================================================
!$OMP PARALLEL DO PRIVATE(jw,iw)
  do iw=1,NS
    do jw=1,NS
      this%INV_NORMAL_DIAGP(iw,jw) = 0.0_DP
      this%INV_NORMAL_DIAGN(iw,jw) = 0.0_DP
      TMPP(iw,jw) = 0.0_DP
      TMPN(iw,jw) = 0.0_DP
    enddo
    this%INV_NORMAL_DIAGP(iw,iw) = 1.0_DP
    this%INV_NORMAL_DIAGN(iw,iw) = 1.0_DP
  enddo
!$OMP PARALLEL DO PRIVATE(jw,iw,kw)
  do jw=1,NS
  do iw=1,NS
    TMPP(iw,jw) = 4.0_DP*delta(iw,jw)
    TMPN(iw,jw) = 4.0_DP*delta(iw,jw)
    do kw=1,NS
      TMPP(iw,jw) = TMPP(iw,jw) + this%KP(kw,iw)*this%KP(kw,jw)
      TMPN(iw,jw) = TMPN(iw,jw) + this%KN(kw,iw)*this%KN(kw,jw)
    enddo
  enddo
  enddo
  call LA_GESV(TMPP,this%INV_NORMAL_DIAGP,INFO=INFO)
  if (INFO /= 0) then
    write(str,'("Error stop at LA_GESV for INV_NORMAL_DIAGP: INFO =",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(str)
  endif
  call LA_GESV(TMPN,this%INV_NORMAL_DIAGN,INFO=INFO)
  if (INFO /= 0) then
    write(str,'("Error stop at LA_GESV for INV_NORMAL_DIAGN: INFO =",I4,1X,A,I5)')INFO,__FILE__,__LINE__
    call error_stop(str)
  endif

  return
end subroutine

function delta(i,j) result(id)
  implicit none
  integer, intent(in) :: i,j
  integer :: id
  if (i==j) then
    id = 1
  else
    id = 0
  endif
  return
end function

subroutine assign_mult_invmprec_dwq(this,dq,q,dagger)
!
! Multiply inverse chiral symmetry improvement term, MB = (B + C M5)
!
!    dq := MB \ q
! or
!    dq := MB^dag \ q
!
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(in)    ::  q
  logical,       optional, intent(in)    :: dagger  ! if present, MD^dagger is multiplied.
  logical :: is_dagger
  type(su3fv_spinor) :: yt,yy(this%NS)
  integer :: NS
  integer :: ix,iy,iz,it,iw
  integer :: kw

  NS = this%NS
  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif

  if (.not.is_dagger) then
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,kw,yy,yt)
    do ix=1,NX
    do iy=1,NY         
    do iz=1,NZ
    do it=1,NT

      do iw=1,NS
        yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
        yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
      enddo

      do iw=1,NS
        yt%y(:,:) = Z0
        do kw=1,NS
          yt%y(:,1:2) = yt%y(:,1:2) + this%INVMBP(iw,kw) * yy(kw)%y(:,1:2)
          yt%y(:,3:4) = yt%y(:,3:4) + this%INVMBN(iw,kw) * yy(kw)%y(:,3:4)
        enddo
        dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
        dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
      enddo

    enddo
    enddo
    enddo
    enddo
  else
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,iw,kw,yy,yt)
    do ix=1,NX
    do iy=1,NY         
    do iz=1,NZ
    do it=1,NT

      do iw=1,NS
        yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
        yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
      enddo

      do iw=1,NS
        yt%y(:,:) = Z0
        do kw=1,NS
          yt%y(:,1:2) = yt%y(:,1:2) + this%INVMBP(kw,iw) * yy(kw)%y(:,1:2)
          yt%y(:,3:4) = yt%y(:,3:4) + this%INVMBN(kw,iw) * yy(kw)%y(:,3:4)
        enddo
        dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
        dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
      enddo

    enddo
    enddo
    enddo
    enddo
  endif

  return
end subroutine

subroutine assign_mult_mprecd_dwq(this,dq,q,u)
!
! Assign and multiply MB preconditioned Domainwall operator
!
!  dq = (D + K) q
!
!  where  K = 4 - M + MA / MB,  MA = 1 - M5,  MB = B - C M5
!
  use counter_class
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(su3fv_spinor) :: yt,yw,yy(this%NS)
  complex(DP) :: w(COL,SPIN/2), y(COL,SPIN/2)
  integer :: NS
  integer :: ix,iy,iz,it,itb,itbup,itbdn,ieoxyz,ieo,ioe
  integer :: iw,iwup,iwdn,kup,kdn
  integer :: kw
  real(DP) :: fup,fdn,mq,dwfac, MM

  NS = this%NS
  if ( (q%NS /= NS) .or. (.not. is_same_size(q,dq)) ) then
    call error_stop("5-dim size of q,dq is wrong in assign_mult_mprec_dwq.")
  endif

  call copy_boundary(q)

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,ieo,ioe,ieoxyz,itb,itbup,itbdn,iw,yt,w,y,yw,kw,yy)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT
    itb = it/2
    ieo = mod(ipeo+it+iz+iy+ix,2)
    ioe = 1-ieo
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    itbup=   +ieoxyz 
    itbdn=-(1-ieoxyz)

    do iw=1,NS
      yy(iw)%y(:,1:2) = q%s(kw,it,iz,iy,ix)%y(:,1:2) - q%s(kw,it,iz,iy,ix)%y(:,3:4)
      yy(iw)%y(:,3:4) = q%s(kw,it,iz,iy,ix)%y(:,1:2) + q%s(kw,it,iz,iy,ix)%y(:,3:4)
    enddo

    do iw=1,NS
      yt%y(:,:) = Z0
      do kw=1,NS
        yt%y(:,1:2) = yt%y(:,1:2) + this%KP(iw,kw)* yy(kw)%y(:,1:2)
        yt%y(:,3:4) = yt%y(:,3:4) + this%KN(iw,kw)* yy(kw)%y(:,3:4)
      enddo
      yw%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4))*0.5_DP
      yw%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4))*0.5_DP

#include "mult_dw_hop4_kernel_normal.h90"

      dq%s(iw,it,iz,iy,ix)%y(:,:) = yw%y(:,:) - 0.5_DP*yt%y(:,:)

    enddo

  enddo
  enddo
  enddo
  enddo
  call inc(mult_iter,NS*2)

  return
end subroutine

subroutine assign_mult_mprecd_dagger_dwq(this,dq,q,u)
!
! Assign and multiply MB preconditioned Domainwall dagger operator
!
!  dq = (D + K)^dag q
!
!  where  K = 4 - M + MA / MB,  MA = 1 - M5,  MB = B - C M5
!
  use counter_class
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(su3fv_spinor) :: yt,yw,yy(this%NS)
  complex(DP) :: w(COL,SPIN/2), y(COL,SPIN/2)
  integer :: NS
  integer :: ix,iy,iz,it,itb,itbup,itbdn,ieoxyz,ieo,ioe
  integer :: iw,iwup,iwdn,kup,kdn
  integer :: kw
  real(DP) :: fup,fdn,mq,dwfac,MM

  NS = this%NS
  if ( (q%NS /= NS) .or. (.not. is_same_size(q,dq)) ) then
    call error_stop("5-dim size of q,dq is wrong in assign_mult_mprec_dagger_dwq.")
  endif

  call copy_boundary(q)

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,it,ieo,ioe,ieoxyz,itb,itbup,itbdn,iw,yt,w,y,yw,kw,yy)
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
  do it=1,NT
    itb = it/2
    ieo = mod(ipeo+it+iz+iy+ix,2)
    ioe = 1-ieo
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    itbup=   +ieoxyz 
    itbdn=-(1-ieoxyz)

    do iw=1,NS
      yy(iw)%y(:,1:2) = q%s(kw,it,iz,iy,ix)%y(:,1:2) - q%s(kw,it,iz,iy,ix)%y(:,3:4)
      yy(iw)%y(:,3:4) = q%s(kw,it,iz,iy,ix)%y(:,1:2) + q%s(kw,it,iz,iy,ix)%y(:,3:4)
    enddo

    do iw=1,NS
      yt%y(:,:) = Z0
      do kw=1,NS
        yt%y(:,1:2) = yt%y(:,1:2) + this%KP(kw,iw)* yy(kw)%y(:,1:2)
        yt%y(:,3:4) = yt%y(:,3:4) + this%KN(kw,iw)* yy(kw)%y(:,3:4)
      enddo
      yw%y(:,1:2) = (+ yt%y(:,1:2) +  yt%y(:,3:4))*0.5_DP
      yw%y(:,3:4) = (- yt%y(:,1:2) +  yt%y(:,3:4))*0.5_DP

#include "mult_dw_hop4_dagger_kernel.h90"

      dq%s(iw,it,iz,iy,ix)%y(:,:) = yw%y(:,:) - 0.5_DP*yt%y(:,:)

    enddo
  enddo
  enddo
  enddo
  enddo
  call inc(mult_iter,NS*2)

  return
end subroutine


subroutine assign_mult_eoprec_dwq(this,ieo,dq,q,vu,dagger)
!
!  Assign (4D) even/odd site preconditioned (MB preconditioned) Domainwall operator
!
!         v
!  dqe =  Dee qe
!
! where
!  v
!  Dee =  1 -  Dee \ Deo Doo \ Doe
!
! or its dagger version
!
!         v
!  dqe =  Dee^dag qe
!
! where
!  v
!  Dee^dag =  1 -  Deo^dag Doo^dag \ Doe^dag Dee^dag \
!
  use counter_class
  implicit none
  class(quark_domainwall), intent(in)    :: this
  integer,                 intent(in)    :: ieo           ! even/odd ness 
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_dw_gluon_wg),   intent(in) :: vu
  logical, optional, intent(in) :: dagger
  type(field_dw_quark_wg), allocatable :: v
  integer :: NS,ioe
  logical :: is_dagger

  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif
  ioe = 1-ieo
  NS = this%NS
  allocate(v)
  call new(v,NS)

  if (.not.is_dagger) then
    call     assign_mult_eo_dwq(this,ioe,dq,q,vu)   !  dqo := Doe qe
    call assign_inv_mult_ee_dwq(this,ioe,v,dq)      !   vo := Doo \ dqo
    call     assign_mult_eo_dwq(this,ieo,dq,v,vu)   !  dqe := Deo vo
    call assign_inv_mult_ee_dwq(this,ieo,v,dq)      !   ve := Dee \ dqe
    call assign_sub(dq,q,v,ieo)                     !  dqe := qe - ve
  else
    call assign_inv_mult_ee_dwq(this,ieo,dq,q,dagger=.true.)      !  dqe := Dee' \ qe
    call     assign_mult_eo_dwq(this,ioe,v,dq,vu,dagger=.true.)   !   vo := Doe' dqe
    call assign_inv_mult_ee_dwq(this,ioe,dq,v,dagger=.true.)      !  dqo := Doo' \ vo
    call     assign_mult_eo_dwq(this,ieo,v,dq,vu,dagger=.true.)   !   ve := Deo' dqo
    call assign_sub(dq,q,v,ieo)                                   !  dqe := qe - ve
  endif

  call delete(v)
  deallocate(v)

  call inc(mult_iter,NS*2)

  return
end subroutine

subroutine assign_inv_mult_ee_dwq(this,ieo,dq,q,dagger)
!
! even-even/odd-odd site component inverse multiplication
!
! dq = K \ q
!
! or
!
! dq = (K^dag \ q
!
! where M : domainwall height
!
!     K = 4 - M + (1 - M5)/(B + C M5),  M5: 5th-direction hopping matrix/2
!
  implicit none
  class(quark_domainwall), intent(in)    :: this
  integer,                 intent(in)    :: ieo
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(in)    ::  q
  logical, optional,       intent(in)    :: dagger
  type(su3fv_spinor) :: yt,yy(this%NS)
  integer :: ix,iy,iz,itb,ieoxyz,it,iw,kw,NS
  logical :: is_dagger

  NS = this%NS

  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif

  if (.not.is_dagger) then

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,it,iw,kw,yt,yy)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
        it = itb*2+ieoxyz

        do iw=1,NS
          yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
          yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
        enddo
        do iw=1,NS
          yt%y(:,:) = Z0
          do kw=1,NS
            yt%y(:,1:2) = yt%y(:,1:2) + this%INVDIAGP(iw,kw) * yy(kw)%y(:,1:2) 
            yt%y(:,3:4) = yt%y(:,3:4) + this%INVDIAGN(iw,kw) * yy(kw)%y(:,3:4) 
          enddo
          dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
          dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
        enddo

      enddo
    enddo
    enddo
    enddo

  else

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,it,iw,kw,yt,yy)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
        it = itb*2+ieoxyz

        do iw=1,NS
          yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
          yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
        enddo
        do iw=1,NS
          yt%y(:,:) = Z0
          do kw=1,NS
            yt%y(:,1:2) = yt%y(:,1:2) + this%INVDIAGP(kw,iw) * yy(kw)%y(:,1:2) 
            yt%y(:,3:4) = yt%y(:,3:4) + this%INVDIAGN(kw,iw) * yy(kw)%y(:,3:4) 
          enddo
          dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
          dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4) )*0.5_DP
        enddo

      enddo
    enddo
    enddo
    enddo
  endif

  return
end subroutine

subroutine assign_mult_ee_dwq(this,ieo,dq,q,dagger)
!
! even-even/odd-odd site component multiplication
!
!   dq = K q
! or
!   dq = K' q
!
! where M : domainwall height
!
!  K = 4 - M + (1 - M5)/(B + C M5)
!
  implicit none
  class(quark_domainwall), intent(in)    :: this
  integer,                 intent(in)    :: ieo
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(in)    :: q
  logical, optional,       intent(in)    :: dagger
  type(su3fv_spinor) :: yt,yy(this%NS)
  real(DP) :: dwfac
  integer :: ix,iy,iz,itb,ieoxyz,it,iw,kw,NS
  logical :: is_dagger

  NS = this%NS

  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif

  if (.not.is_dagger) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,it,iw,kw,yt,yy)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
        it = itb*2+ieoxyz

        do iw=1,NS
          yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
          yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
        enddo
        do iw=1,NS
          yt%y(:,:)=Z0
          do kw=1,NS
            yt%y(:,1:2) = yt%y(:,1:2) + this%KP(iw,kw) * yy(kw)%y(:,1:2)
            yt%y(:,3:4) = yt%y(:,3:4) + this%KN(iw,kw) * yy(kw)%y(:,3:4)
          enddo
          dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4))*0.5_DP
          dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4))*0.5_DP
        enddo

      enddo
    enddo
    enddo
    enddo
  else
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,it,iw,kw,yt,yy)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
        it = itb*2+ieoxyz

        do iw=1,NS
          yy(iw)%y(:,1:2) = q%s(iw,it,iz,iy,ix)%y(:,1:2) - q%s(iw,it,iz,iy,ix)%y(:,3:4)
          yy(iw)%y(:,3:4) = q%s(iw,it,iz,iy,ix)%y(:,1:2) + q%s(iw,it,iz,iy,ix)%y(:,3:4)
        enddo
        do iw=1,NS
          yt%y(:,:)=Z0
          do kw=1,NS
            yt%y(:,1:2) = yt%y(:,1:2) + this%KP(kw,iw) * yy(kw)%y(:,1:2)
            yt%y(:,3:4) = yt%y(:,3:4) + this%KN(kw,iw) * yy(kw)%y(:,3:4)
          enddo
          dq%s(iw,it,iz,iy,ix)%y(:,1:2) = (+ yt%y(:,1:2) + yt%y(:,3:4))*0.5_DP
          dq%s(iw,it,iz,iy,ix)%y(:,3:4) = (- yt%y(:,1:2) + yt%y(:,3:4))*0.5_DP
        enddo

      enddo
    enddo
    enddo
    enddo
  endif

  return
end subroutine

subroutine assign_mult_eo_dwq(this,ieo,dq,q,vu,dagger)
!
! even-odd/odd-even site component multiplication
!
! dq(iw,ne) = - (1/2) sum_{mu=1,4}[ (1-gamma_mu) U_mu(ne)          q(iw,ne+mu) 
!                                 + (1+gamma_mu) U_mu(ne-mu)^{dag} q(iw,ne-mu) ]
!
! ne : 4-D even/odd site
! iw : 5th direction
!
  implicit none
  class(quark_domainwall),  intent(in)    :: this
  integer,                  intent(in)    :: ieo
  type(field_dw_quark_wg),  intent(inout) :: dq
  type(field_dw_quark_wg),  intent(inout) ::  q
  type(vfield_dw_gluon_wg), intent(in)    :: vu
  logical, optional,        intent(in)    :: dagger
  integer :: ix,iy,iz,itb,ieoxyz,it,iw,NS,ioe
  complex(DP) :: y(COL,SPIN/2),w(COL,SPIN/2)
  type(su3fv_spinor) :: yt
  logical :: is_dagger

  NS = this%NS

  if (present(dagger)) then
    is_dagger = .true.
  else
    is_dagger = .false.
  endif

  call copy_boundary(q)

  if (.not.is_dagger) then
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,it,iw,y,w,yt)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
      it = itb*2+ieoxyz
      do iw=1,NS
#include "mult_eo.h90"
      enddo
      enddo
    enddo
    enddo
    enddo
  else
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,it,iw,y,w,yt)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
      it = itb*2+ieoxyz
      do iw=1,NS
#include "mult_eo_dag.h90"
      enddo
      enddo
    enddo
    enddo
    enddo
  endif

  return
end subroutine

subroutine make_midterm_eoprec_lowprojd_dwq(this,ieo,LV,GGe)
!
! Compute DWF chiral-symmetry improvement term using low-modes.
!
!
! GGe = (1 + X We^dag gamma5 Dee \ We)^(-1) X
!
  use f95_lapack, only : LA_GESV
  implicit none
  class(quark_domainwall), intent(in)  :: this
  integer,                 intent(in)  :: ieo
  type(dwf_low_modes),     intent(in)  :: LV
  complex(DP), intent(out) :: GGe(LV%NEV,this%NS,LV%NEV,this%NS)

  type(field_dw_quark_wg),  allocatable :: w,v
  complex(DP), allocatable :: MID(:,:),INVMID(:,:)
  complex(DP) :: ctmp
  integer :: NS,NEV,iev,jev,jw,kev,kw,ic,is
  integer :: ix,iy,iz,it,iw,ith,ieoxyz

  NS  = this%NS
  NEV = LV%NEV

  allocate(v,w)
  call new(v,NS)
  call new(w,NS)

  !
  !  MID = We^dag gamma5 Dee \ We
  !
  allocate(MID(NS*NEV,NS*NEV))
  MID(:,:) = Z0

  do jev=1,NEV
  do jw=1,NS

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ieoxyz,ith,iw)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz
        it = ith*2 + ieoxyz
        do iw=1,jw-1
          v%s(iw,it,iz,iy,ix)%y(:,:) = Z0
        enddo
        v%s(jw,it,iz,iy,ix)%y(:,:) = LV%W%s(jev,it,iz,iy,ix)%y(:,:)
        do iw=jw+1,NS
          v%s(iw,it,iz,iy,ix)%y(:,:) = Z0
        enddo
      enddo
    enddo
    enddo
    enddo

    call assign_inv_mult_ee_dwq(this,ieo,w,v)      !  we := Dee \ ve(jw)

    do iev=1,NEV
    do iw=1,NS
      ctmp = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ieoxyz,ith,ic,is) REDUCTION(+:ctmp)
      do ix=1,NX
      do iy=1,NY
      do iz=1,NZ
        ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
        do ith=1-ieoxyz,NTH-ieoxyz
          it = ith*2 + ieoxyz
          do is=1,SPIN
          do ic=1,COL
            ctmp = ctmp + conjg(LV%W%s(iev,it,iz,iy,ix)%y(ic,    is))  &
  &                              * w%s( iw,it,iz,iy,ix)%y(ic,igamma(is,5)) 
          enddo
          enddo
        enddo
      enddo
      enddo
      enddo
#ifndef _singlePU
      call comlib_sumcast(ctmp)
#endif
      MID(iev+NEV*(iw-1),jev+NEV*(jw-1)) = ctmp
    enddo
    enddo

  enddo
  enddo
  call delete(w)
  call delete(v)
  deallocate(v,w)

  allocate(INVMID(NS*NEV,NS*NEV))

  INVMID(:,:) = Z0
!$OMP PARALLEL DO PRIVATE(jw,jev,iw,iev,kev)
  do jw=1,NS
  do jev=1,NEV
  do iw=1,NS
  do iev=1,NEV
    do kev=1,NEV
      INVMID(iev+NEV*(iw-1),jev+NEV*(jw-1)) = INVMID(iev+NEV*(iw-1),jev+NEV*(jw-1)) &
  &                            + LV%XX(iev,kev)* MID(kev+NEV*(iw-1),jev+NEV*(jw-1)) 
    enddo
  enddo
  enddo
  enddo
  enddo

  !
  !  MID = 1 + X We^dag gamma5 Dee \ We
  !
  MID(:,:) = INVMID(:,:)
  do iw=1,NEV*NS
    MID(iw,iw) = MID(iw,iw) + Z1
  enddo

  !
  !  GGe = (1 + X We^dag gamma5 Dee \ We)^(-1) X = MID \ X
  !
  INVMID(:,:) = Z0
!$OMP PARALLEL DO PRIVATE(iw,jev,iev)
  do iw=1,NS
  do jev=1,NEV
  do iev=1,NEV
    INVMID(iev+NEV*(iw-1),jev+NEV*(iw-1)) = LV%XX(iev,jev)
  enddo
  enddo
  enddo

  call LA_GESV(MID,INVMID)

!$OMP PARALLEL DO PRIVATE(jw,jev,iw,iev)
  do jw=1,NS
  do jev=1,NEV
  do iw=1,NS
  do iev=1,NEV
    GGe(iev,iw,jev,jw) = INVMID(iev+NEV*(iw-1),jev+NEV*(jw-1))
  enddo
  enddo
  enddo
  enddo
  deallocate(MID,INVMID)

  return
end subroutine

subroutine set_eoprec_lowprojd_source_dwq(this,ieo,LV,GGo,bb,b,vu)
!
!  Compute even/odd site-preconditioned source vector with low-mode improvement.
!
!  bbe = Dee \ ( be - Deo Doo \ bo - ( We - Deo Doo \ Wo) GGo Wo^dag gamma5 Doo \ bo )
!
  implicit none
  class(quark_domainwall),  intent(in)    :: this
  integer,                  intent(in)    :: ieo
  type(dwf_low_modes),      intent(in)    :: LV
  complex(DP),              intent(in)    :: GGo(LV%NEV,this%NS,LV%NEV,this%NS)
  type(field_dw_quark_wg),  intent(inout) :: bb 
  type(field_dw_quark_wg),  intent(in)    ::  b
  type(vfield_dw_gluon_wg), intent(in)    :: vu

  integer :: ioe,keo
  type(field_dw_quark_wg), allocatable :: v,w
  complex(DP), allocatable :: cf1(:,:),cf2(:,:)
  complex(DP) :: ctmp(LV%NEV*this%NS)
  integer :: NS,NEV,iev,jev,jw,kev,kw,ic,is
  integer :: ix,iy,iz,it,iw,ith,ieoxyz

  ioe = 1-ieo
  NS  = this%NS
  NEV = LV%NEV

  allocate(v,w)
  call new(v,NS)
  call new(w,NS)

  allocate(cf1(NEV,NS))
  allocate(cf2(NEV,NS))

  call assign_inv_mult_ee_dwq(this,ioe,v,b)      !  vo := Doo \ bo

  !
  !  cf1(iev,iw) = Wo(iev)^dag gamma_5 vo(iw)
  !  cf2(iev,iw) = sum_kev,kw GGo(iev,iw,kev,kw) * cf1(kev,kw)
  !
  cf1(:,:) = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ieoxyz,ith,ic,is,iw,iev) REDUCTION(+:cf1)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ioe,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      it = ith*2 + ieoxyz
      do iw=1,NS
      do iev=1,NEV
        do is=1,SPIN
        do ic=1,COL
          cf1(iev,iw) = cf1(iev,iw) + CONJG(LV%W%s(iev,it,iz,iy,ix)%y(ic,       is ))  &
  &                                           *v%s( iw,it,iz,iy,ix)%y(ic,igamma(is,5))
        enddo
        enddo
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo
#ifndef _singlePU
  ctmp(:) = RESHAPE(cf1(:,:),[1])
  call comlib_sumcast(ctmp)
  cf1(:,:) = RESHAPE(ctmp(:),[NEV,NS])
#endif

!$OMP PARALLEL DO PRIVATE(iw,iev,kw,kev)
  do iw=1,NS
  do iev=1,NEV
    cf2(iev,iw) = Z0
    do kw=1,NS
    do kev=1,NEV
      cf2(iev,iw) = cf2(iev,iw) + GGo(iev,iw,kev,kw)*cf1(kev,kw)
    enddo
    enddo
  enddo
  enddo

  !
  !  v(iw) = b(iw) - sum_iev W(iev) cf2(iev,iw)
  !
  call assign(v,b)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,iw,iev)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do it=1,NT
    do iw=1,NS
    do iev=1,NEV
      v%s(iw,it,iz,iy,ix)%y(:,:) = v%s( iw,it,iz,iy,ix)%y(:,:) &
 &                            - LV%W%s(iev,it,iz,iy,ix)%y(:,:)*cf2(iev,iw)
    enddo
    enddo
  enddo
  enddo
  enddo
  enddo

  !
  !  ve := ve - Deo Doo \ vo
  !
  call assign_inv_mult_ee_dwq(this,ioe, w,v)     !  wo := Doo \ vo
  call     assign_mult_eo_dwq(this,ieo,bb,w,vu)  ! bbe := Deo wo
  call accum_sub(v,bb,ieo)                       !  ve := ve - bbe

  !
  !  bbe := Dee \ ve
  !
  call assign_inv_mult_ee_dwq(this,ieo,bb,v)

  call delete(v)
  call delete(w)
  deallocate(v,w)
  deallocate(cf1,cf2)

  return
end subroutine

subroutine get_eoprec_lowprojd_sol_dwq(this,ieo,LV,GGo,dq,b,vu)
!
! Compute odd/even part solution from the even/odd part solution solved from 
! the even/odd site-preconditioned equation.
!
! dqo = Doo \ ( bbo - Wo GGo (We^dag, Wo^dag)gamma5(ye, Doo \ bbo))
!
! bbo = bo - Doe dqe
!
  implicit none
  class(quark_domainwall),  intent(in)    :: this
  integer,                  intent(in)    :: ieo
  type(dwf_low_modes),      intent(in)    :: LV
  complex(DP),              intent(in)    :: GGo(LV%NEV,this%NS,LV%NEV,this%NS)
  type(field_dw_quark_wg),  intent(inout) :: dq
  type(field_dw_quark_wg),  intent(in)    ::  b
  type(vfield_dw_gluon_wg), intent(in)    :: vu

  integer :: ioe,keo
  type(field_dw_quark_wg), allocatable :: v,w
  complex(DP), allocatable :: cf1(:,:),cf2(:,:)
  complex(DP) :: ctmp(LV%NEV*this%NS)
  integer :: NS,NEV,iev,jev,jw,kev,kw,ic,is
  integer :: ix,iy,iz,it,iw,ith,ieoxyz

  ioe = 1-ieo
  NS  = this%NS
  NEV = LV%NEV

  allocate(v,w)
  call new(v,NS)
  call new(w,NS)

  allocate(cf1(NEV,NS))
  allocate(cf2(NEV,NS))

  !
  ! wo :=  bo - Doe dqe
  !
  call assign(w,b,ioe)                           !  wo := bo
  call assign_mult_eo_dwq(this,ioe,v,dq,vu)      !  vo := Doe dqe
  call accum_sub(w,v,ioe)                        !  wo := wo - vo

  !
  ! ve := dqe
  ! vo := Doo \ wo
  !
  !  v = | dqe                  |
  !      | Doo \ (bo - Doe dqe) | 
  !
  call assign(v,dq,ieo)                          !  ve := dqe
  call assign_inv_mult_ee_dwq(this,ioe,v,w)      !  vo := Doo \ wo

  !
  !  cf1(iev,iw) = W(iev)^dag gamma_5 v(iw)
  !  cf2(iev,iw) = sum_kev,kw GGo(iev,iw,kev,kw) * cf1(kev,kw)
  !
  cf1(:,:) = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ic,is,iw,iev) REDUCTION(+:cf1)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do it=1,NT
    do iw=1,NS
    do iev=1,NEV
      do is=1,SPIN
      do ic=1,COL
        cf1(iev,iw) = cf1(iev,iw) + CONJG(LV%W%s(iev,it,iz,iy,ix)%y(ic,       is ))  &
  &                                         *v%s( iw,it,iz,iy,ix)%y(ic,igamma(is,5))
      enddo
      enddo
    enddo
    enddo
  enddo
  enddo
  enddo
  enddo
#ifndef _singlePU
  ctmp(:) = RESHAPE(cf1(:,:),[1])
  call comlib_sumcast(ctmp)
  cf1(:,:) = RESHAPE(ctmp(:),[NEV,NS])
#endif

!$OMP PARALLEL DO PRIVATE(iw,iev,kw,kev)
  do iw=1,NS
  do iev=1,NEV
    cf2(iev,iw) = Z0
    do kw=1,NS
    do kev=1,NEV
      cf2(iev,iw) = cf2(iev,iw) + GGo(iev,iw,kev,kw)*cf1(kev,kw)
    enddo
    enddo
  enddo
  enddo

  !
  !  vo(iw) = wo(iw) - sum_iev Wo(iev) cf2(iev,iw)
  !
  call assign(v,w,ioe)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ieoxyz,ith,iw,iev)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ioe,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      it = ith*2 + ieoxyz
      do iw=1,NS
      do iev=1,NEV
        v%s(iw,it,iz,iy,ix)%y(:,:) = v%s( iw,it,iz,iy,ix)%y(:,:) &
 &                              - LV%W%s(iev,it,iz,iy,ix)%y(:,:)*cf2(iev,iw)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  !
  !  dqo := Doo \ vo
  !
  call assign_inv_mult_ee_dwq(this,ioe,dq,v)

  call delete(v)
  call delete(w)
  deallocate(v,w)
  deallocate(cf1,cf2)

  return
end subroutine
  
subroutine assign_mult_eoprec_lowprojd_dwq(this,ieo,LV,GGo,dq,q,vu,dagger)
!
! Assign and multiply even/odd-site preconditioned lowmode improved DWF operator
!
!        v
!  dqe = Dee qe + Dee \ (We - Deo Doo \ Wo) GGo (We^dag,Wo^dag) gamma5 (qe, -Doo \ Doe qe)
!
! or
!        v
!  dqe = Dee^dag qe
!      + (gamma5 We - Deo^dag Doo^dag \ gamma5 Wo) GGo^dag 
!                               (We^dag,Wo^dag) (Dee^dag \ qe, -Doo^dag \ Doe^dag Dee^dag \ qe)
!
  implicit none
  class(quark_domainwall),  intent(in) :: this
  integer,                  intent(in) :: ieo
  type(dwf_low_modes),      intent(in) :: LV
  complex(DP),              intent(in) :: GGo(LV%NEV,this%NS,LV%NEV,this%NS)
  type(field_dw_quark_wg),  intent(inout) :: dq
  type(field_dw_quark_wg),  intent(inout) ::  q
  type(vfield_dw_gluon_wg), intent(in)    :: vu
  logical, optional, intent(in) :: dagger

  integer :: ioe,keo
  type(field_dw_quark_wg), allocatable :: v,w
  complex(DP), allocatable :: cf1(:,:),cf2(:,:)
  complex(DP) :: ctmp(LV%NEV*this%NS)
  integer :: NS,NEV,iev,jev,jw,kev,kw
  integer :: ix,iy,iz,it,iw,ith,ieoxyz,ic,is
  logical :: is_dagger

  if (present(dagger)) then
    is_dagger=.true.
  else
    is_dagger=.false.
  endif

  ioe = 1-ieo
  NS  = this%NS
  NEV = LV%NEV

  allocate(v,w)
  call new(v,NS)
  call new(w,NS)

  allocate(cf1(NEV,NS))
  allocate(cf2(NEV,NS))

  if (.not.is_dagger) then

    !
    !  v := (qe,- Doo \ Doe qe)
    !
    call assign(v,q,ieo)                           !  ve := qe
    call     assign_mult_eo_dwq(this,ioe,w,q,vu)   !  wo := Doe qe
    call assign_inv_mult_ee_dwq(this,ioe,v,w)      !  vo := Doo \ wo
    call accum_mult(v,-1.0_DP,ioe)                 !  vo := - vo

    !
    !  cf1(iev,iw) = W(iev)^dag gamma_5 v(iw)
    !  cf2(iev,iw) = sum_kev,kw GGo(iev,iw,kev,kw) * cf1(kev,kw)
    !
    cf1(:,:) = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ic,is,iw,iev) REDUCTION(+:cf1)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      do iw=1,NS
      do iev=1,NEV
        do is=1,SPIN
        do ic=1,COL
          cf1(iev,iw) = cf1(iev,iw) + CONJG(LV%W%s(iev,it,iz,iy,ix)%y(ic,       is ))    &
  &                                          * v%s( iw,it,iz,iy,ix)%y(ic,igamma(is,5))
        enddo
        enddo
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo
#ifndef _singlePU
    ctmp(:) = RESHAPE(cf1(:,:),[1])
    call comlib_sumcast(ctmp)
    cf1(:,:) = RESHAPE(ctmp(:),[NEV,NS])
#endif

!$OMP PARALLEL DO PRIVATE(iw,iev,kw,kev)
    do iw=1,NS
    do iev=1,NEV
      cf2(iev,iw) = Z0
      do kw=1,NS
      do kev=1,NEV
        cf2(iev,iw) = cf2(iev,iw) + GGo(iev,iw,kev,kw)*cf1(kev,kw)
      enddo
      enddo
    enddo
    enddo

    !
    ! v(iw) = sum_iev W(iev) cf2(iev,iw)
    !
    call clear(v)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,iw,iev)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      do iw=1,NS
      do iev=1,NEV
        v%s(iw,it,iz,iy,ix)%y(:,:) = v%s( iw,it,iz,iy,ix)%y(:,:) &
 &                              + LV%W%s(iev,it,iz,iy,ix)%y(:,:)*cf2(iev,iw)
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo

    !
    !  we = Dee \ ( ve - Deo Doo \ vo )
    !
    call assign_inv_mult_ee_dwq(this,ioe,w,v)      !  wo := Doo \ vo
    call     assign_mult_eo_dwq(this,ieo,dq,w,vu)  ! dqe := Deo wo
    call accum_sub(v,dq,ieo)                       !  ve := ve - dqe = ve - Deo Doo \ vo
    call assign_inv_mult_ee_dwq(this,ieo,w,v)      !  we := Dee \ ve

    !       v
    ! dq <= Ddw qe + w
    !
    call assign_mult_eoprec_dwq(this,ieo,dq,q,vu)
    call accum_add(dq,w,ieo)

  else

    !
    !  v := (Dee^dag \ qe,- Doo^dag \ Doe^dag Dee^dag \ qe)
    !
    call assign_inv_mult_ee_dwq(this,ieo,v,q,   dagger=.true.)   !  ve := Dee^dag \ qe
    call     assign_mult_eo_dwq(this,ioe,w,v,vu,dagger=.true.)   !  wo := Doe^dag ve
    call assign_inv_mult_ee_dwq(this,ioe,v,w,   dagger=.true.)   !  vo := Doo^dag \ wo
    call accum_mult(v,-1.0_DP,ioe)                               !  vo := - vo

    !
    !  cf1(iev,iw) = W(iev)^dag v(iw)
    !  cf2(iev,iw) = sum_kev,kw GGo(kev,kw,iev,iw)^* * cf1(kev,kw)
    !
    cf1(:,:) = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ic,is,iw,iev) REDUCTION(+:cf1)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      do iw=1,NS
      do iev=1,NEV
        do is=1,SPIN
        do ic=1,COL
          cf1(iev,iw) = cf1(iev,iw) + CONJG(LV%W%s(iev,it,iz,iy,ix)%y(ic,is))     &
  &                                          * v%s( iw,it,iz,iy,ix)%y(ic,is)
        enddo
        enddo
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo
#ifndef _singlePU
    ctmp(:) = RESHAPE(cf1(:,:),[1])
    call comlib_sumcast(ctmp)
    cf1(:,:) = RESHAPE(ctmp(:),[NEV,NS])
#endif

!$OMP PARALLEL DO PRIVATE(iw,iev,kw,kev)
    do iw=1,NS
    do iev=1,NEV
      cf2(iev,iw) = Z0
      do kw=1,NS
      do kev=1,NEV
        cf2(iev,iw) = cf2(iev,iw) + CONJG(GGo(kev,kw,iev,iw)) * cf1(kev,kw)
      enddo
      enddo
    enddo
    enddo

    !
    ! v(iw) = sum_iev gamma_5 W(iev) cf2(iev,iw) 
    !
    call clear(v)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,iw,iev,ic,is)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      do iw=1,NS
      do iev=1,NEV
        do is=1,SPIN
        do ic=1,COL
          v%s(iw,it,iz,iy,ix)%y(ic,is) = v%s( iw,it,iz,iy,ix)%y(ic,       is) &
 &                                  + LV%W%s(iev,it,iz,iy,ix)%y(ic,igamma(is,5))*cf2(iev,iw)
        enddo
        enddo
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo

    !
    ! ve = ve - Deo^dag Doo^dag \ vo
    !
    call assign_inv_mult_ee_dwq(this,ioe, w,v,dagger=.true.)      !  wo := Doo^dag \ vo
    call     assign_mult_eo_dwq(this,ieo,dq,w,vu,dagger=.true.)   ! dqe := Deo^dag wo
    call accum_sub(v,dq,ieo)                                      !  ve := ve - dqe

    !      v
    ! v <= Ddw^dag q + v
    !
    call assign_mult_eoprec_dwq(this,ieo,dq,q,vu,dagger=.true.)
    call accum_add(dq,v,ieo)

  endif

  call delete(v)
  call delete(w)
  deallocate(v,w)
  deallocate(cf1,cf2)

  return
end subroutine

subroutine assign_mult_lowprojd_dwq(this,dq,q,u,LV,dagger)
!
! Assign and multiply Domainwall operator with low-mode improvement
!
! dq <= (Ddw     + W X W^dag gamma_5 (B+C M_5) ) q 
!
! or
!
! dq <= (Ddw^dag + (B + C M_5^dag) gamma_5 W X W^dag) q 
!
! for dagger replace gamma_mu => -gamma_mu.
!
  implicit none
  class(quark_domainwall), intent(in)    :: this
  type(field_dw_quark_wg), intent(inout) :: dq
  type(field_dw_quark_wg), intent(inout) ::  q
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(dwf_low_modes),     intent(in)    :: LV
  logical, optional, intent(in) :: dagger

  type(field_dw_quark_wg), allocatable :: v
  complex(DP), allocatable :: cf1(:,:),cf2(:,:)
  complex(DP) :: ctmp
  integer :: NS,NEV,iev,kev
  integer :: ix,iy,iz,it,iw,ith,ieoxyz,ieo,ioe,ic,is
  logical :: is_dagger

  if (present(dagger)) then
    is_dagger=.true.
  else
    is_dagger=.false.
  endif

  NS  = q%NS
  NEV = LV%NEV
  allocate(v)
  call new(v,NS)
  allocate(cf1(NEV,NS))
  allocate(cf2(NEV,NS))

  if (.not. is_dagger) then
    !
    ! dq <= Ddw + W X W^dag \gamma_5 (B + C M5)
    !

    !
    ! dq <= Ddw q
    !
    call assign_mult_dwf(this,dq,q,u)

    !
    !  v <= (B + C M5) q
    !
    call assign_mult_impterm_dwf(this,v,q)

    !
    !  cf1(iev,iw) = W(iev)^dag gamma_5 v(iw)
    !  cf2(iev,iw) = sum_kev(X(iev,kev) * cf1(kev,iw))
    !
    do iw=1,NS
    do iev=1,NEV
      ctmp = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ic,is) REDUCTION(+:ctmp)
      do ix=1,NX
      do iy=1,NY
      do iz=1,NZ
      do it=1,NT
        do is=1,SPIN
        do ic=1,COL
          ctmp = ctmp + CONJG(LV%W%s(iev,it,iz,iy,ix)%y(ic,       is ))   &
  &                            * v%s( iw,it,iz,iy,ix)%y(ic,igamma(is,5))
        enddo
        enddo
      enddo
      enddo
      enddo
      enddo
#ifndef _singlePU
      call comlib_sumcast(ctmp)
#endif
      cf1(iev,iw) = ctmp
    enddo
    enddo

!   
!    cf2(1:NEV,1:NS) = MATMUL(LV%XX(1:NEV,1:NEV),cf1(1:NEV,1:NS))
!   
!$OMP PARALLEL DO PRIVATE(iw,iev,kev)
    do iw=1,NS
    do iev=1,NEV
      cf2(iev,iw) = Z0
      do kev=1,NEV
        cf2(iev,iw) = cf2(iev,iw) + LV%XX(iev,kev)*cf1(kev,iw)
      enddo
    enddo
    enddo

    !
    ! dq(iw) = dq(iw) + sum_iev W(iev) cf2(iev,iw)
    !
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,iw,iev)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      do iw=1,NS
      do iev=1,NEV
        dq%s(iw,it,iz,iy,ix)%y(:,:) = dq%s( iw,it,iz,iy,ix)%y(:,:) &
 &                                + LV%W%s(iev,it,iz,iy,ix)%y(:,:)*cf2(iev,iw)
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo

  else

    !
    ! dq <= Ddw^dag + (B+C M_5)^dag gamma_5 W X W^dag
    !

    !
    !  cf1(iev,iw) = W(iev)^dag q(iw)
    !  cf2(iev,iw) = sum_kev(X(kev,iev)^* * cf1(kev,iw))
    !
    do iw=1,NS
    do iev=1,NEV
      ctmp = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,ic,is) REDUCTION(+:ctmp)
      do ix=1,NX
      do iy=1,NY
      do iz=1,NZ
      do it=1,NT
        do is=1,SPIN
        do ic=1,COL
          ctmp = ctmp + CONJG(LV%W%s(iev,it,iz,iy,ix)%y(ic,is))  &
  &                            * q%s( iw,it,iz,iy,ix)%y(ic,is)
        enddo
        enddo
      enddo
      enddo
      enddo
      enddo
#ifndef _singlePU
      call comlib_sumcast(ctmp)
#endif
      cf1(iev,iw) = ctmp
    enddo
    enddo

!
!    cf2(1:NEV,1:NS) = MATMUL(CONJG(TRANSPOSE(LV%XX(1:NEV,1:NEV))),cf1(1:NEV,1:NS))
!
!$OMP PARALLEL DO PRIVATE(iw,iev,kev)
    do iw=1,NS
    do iev=1,NEV
      cf2(iev,iw) = Z0
      do kev=1,NEV
        cf2(iev,iw) = cf2(iev,iw) + CONJG(LV%XX(kev,iev))*cf1(kev,iw)
      enddo
    enddo
    enddo

    !
    ! v(iw) = sum_iev (gamma_5 W_(iev)) cf2(iev,iw) = (gamma_5 W X W^dag)(iw)
    !
    call clear(v)
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,it,iw,iev)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
    do it=1,NT
      do iw=1,NS
      do iev=1,NEV
        v%s(iw,it,iz,iy,ix)%y(:,:) = v%s( iw,it,iz,iy,ix)%y(:,:) &
 &                              + LV%W%s(iev,it,iz,iy,ix)%y(:,:)*cf2(iev,iw)
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo

    !
    ! dq <= (B + M_5^dag C) v 
    !     = (B + M_5^dag C) gamma_5 W X W^dag
    !
    call assign_mult_impterm_dwf(this,dq,v,dagger=.true.)

    !
    ! v <= Ddw^dag q
    !
    call assign_mult_dagger_dwq(this,v,q,u)

    !
    ! dq <= dq + v
    !
    call accum_add(dq,v)

  endif

  call delete(v)
  deallocate(v)
  deallocate(cf1,cf2)

  return
end subroutine


subroutine assign_inv_mult_lowprojd_dwq(this,iout,tol,iiter,dy,y,u,LV)
!
! Assign and multiply inverse low-mode improved Domainwall operator
!
! dy <= Ddw \ y
!
! using CGNE solver
!
  use solver_class
  use counter_class
  implicit none
  class(quark_domainwall), intent(inout) :: this
  integer,                 intent(in)    :: iout
  real(DP),                intent(in)    :: tol
  integer,                 intent(inout) :: iiter
  type(field_dw_quark_wg), intent(inout) :: dy
  type(field_dw_quark_wg), intent(inout) ::  y
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(dwf_low_modes),     intent(in)    :: LV

#if _PREC_ == _PREC_EO_
  integer, parameter :: SLV_NSITE=COL*SPIN*NTH*NZ*NY*NX
#else
  integer, parameter :: SLV_NSITE=COL*SPIN*NT*NZ*NY*NX
#endif
#if _PREC_ == _PREC_EO_
  type(vfield_dw_gluon_wg), allocatable :: vu
  complex(DP), allocatable :: GGo(:,:,:,:)
  integer :: ieo,ioe
  type(field_dw_quark_wg),  allocatable :: bb
#endif
  type(field_dw_quark_wg),  allocatable :: w,v
  type(cg_alg),             allocatable :: solver
  type(timer) :: solver_time
  character(CHARLEN) :: str
  real(DP) :: etime,rtmp0,rtmp1
  integer    :: NS,istat,mult_count
  type(counter) :: imult
  call new(solver_time)
  call tic(solver_time)

  NS = this%NS
  if (NS /= y%NS) then
    call error_stop("this%NY /= y%NS, input y is wrong in assign_inv_mult_lowprojd_dwq.")
  endif
  if (NS /= dy%NS) then
    call new(dy,NS)
  endif

  allocate(w,v)
  call new(w,this%NS)
  call new(v,this%NS)

  imult = mult_iter

  !==============================================
  ! CGNE is used to solve
  ! D x = y.
  !
  !  (D' D) x = D' y
  !
  !==============================================

#if _PREC_ == _PREC_NON_
  !==============
  ! set source
  !  w = D' y
  !==============
  call assign_mult_dwf(this,LV,w,y,u,dagger=.true.)

#elif _PREC_ == _PREC_EO_

  ieo = 0
  ioe = 1-ieo
  allocate(vu)
  allocate(GGo(LV%NEV,NS,LV%NEV,NS))
  call conv_f_dwg_wg(u,vu)
  call make_midterm_eoprec_lowprojd_dwq(this,ioe,LV,GGo)
  call   set_eoprec_lowprojd_source_dwq(this,ieo,LV,GGo,w,y,vu)

#endif

  !==============
  ! set solver
  !==============
  allocate(solver)
  call new(solver,NSIZE=SLV_NSITE*NS,mode=MODE_NORMAL    ,max_iter=iiter,tol=tol)
#if _PREC_ == _PREC_EO_
  call pack(w,solver%src_vec(:),ieo)
#else
  call pack(w,solver%src_vec(:))
#endif

  !=======================
  ! start CG Solver
  !=======================
  do 
    call solve(solver)
    istat = get_status(solver)

    select case(istat)
    case (OP_NOP)

      cycle

    case (OP_DO_MATVEC)

      !===================
      ! Multiply  (D'D)
      !===================
#if _PREC_ == _PREC_NON_
      call unpack(solver%src_vec,w)
      call assign_mult_dwf(this,LV,v,w,u)
      call assign_mult_dwf(this,LV,w,v,u,dagger=.true.)
      call pack(w,solver%dst_vec)
#elif _PREC_ == _PREC_EO_
      call unpack(solver%src_vec,w,ieo)
      call assign_mult_eoprec_lowprojd_dwq(this,ieo,LV,GGo,v,w,vu,dagger=.true.)
      call assign_mult_eoprec_lowprojd_dwq(this,ieo,LV,GGo,w,v,vu)
      call pack(w,solver%dst_vec,ieo)
#endif

    case (OP_PRINT_STATUS)
#ifdef _DEBUG
      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)') &
 &                        get_current_iteration(solver),&
 &                            get_residual_norm(solver),mult_count
        write(*   ,'("#",I5," ERR_=",E24.16," MULT_=",I10)') &
 &                        get_current_iteration(solver),&
 &                            get_residual_norm(solver),mult_count
      endif
#endif
      cycle
    case (OP_CONVERGED)

      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",A," iteration converged.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)',advance='no')     &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count
      endif
      exit      

    case (OP_MAXITER_REACHED)

      if (nodeid == 0) then
        mult_count = get_count(mult_iter) - get_count(imult)
        write(iout,'("#",A," iteration does not converge.")')TRIM(get_name(solver))
        write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)',advance='no')     &
 &                          get_current_iteration(solver),&
 &                              get_residual_norm(solver),mult_count
      endif
      write(str,'("solver did not converge.: ",A,I4)')__FILE__,__LINE__
      call error_stop(str)

    case default

      write(str,'("something is wrong in solver.: ",A,I4)')__FILE__,__LINE__
      call error_stop(str)

    end select
  enddo

#if _PREC_ == _PREC_EO_
  call unpack(solver%dst_vec,v,ieo)
  call assign_mult_eoprec_lowprojd_dwq(this,ieo,LV,GGo,w,v,vu,dagger=.true.)
  call get_eoprec_lowprojd_sol_dwq(this,ieo,LV,GGo,w,y,vu)
  call assign_mult_invmprec_dwq(this,dy,w)
  deallocate(vu)
  deallocate(GGo)
#else
  call unpack(solver%dst_vec,dy)
#endif

  call delete(v)
  call delete(w)
  deallocate(w,v)

  iiter = get_current_iteration(solver)

  call delete(solver)
  deallocate(solver)

  !============================================
  ! NOTE: get_elapse contains comlib_sumcast.
  ! This means that get_elapse should not used
  ! in if-(nodeid==0)-then-endif block.
  !============================================
  call toc(solver_time)
  etime = get_elapse(solver_time)  
  if (nodeid == 0) then
    write(iout,'(" ETIME_=",E24.16)')etime
  endif

#ifdef _DEBUG
  allocate(w)
  call new(w,this%NS)
  call assign_mult_dwf(this,w,dy,u,LV)
  call accum_sub(w,y)
  rtmp0 = abs2(w)
  rtmp1 = abs2(y)
  if (nodeid==0) then
    write(iout,'("%",I5," ERR_=",3E24.16)',advance='no') &
 &            iiter,sqrt(rtmp0/rtmp1),sqrt(rtmp0),sqrt(rtmp1)
    write(iout,'(" ETIME_=",E24.16)')etime
    write(*   ,'("%",I5," ERR_=",3E24.16)',advance='no') &
 &            iiter,sqrt(rtmp0/rtmp1),sqrt(rtmp0),sqrt(rtmp1)
    write(*   ,'(" ETIME_=",E24.16)')etime
  endif
  call delete(w)
  deallocate(w)
#endif

  return
end subroutine

#include "4dssor_solver.h90"

end module
